@@ -20,6 +20,7 @@
 #include "hash.h"
 #include "ident.h"
 #include "phase.h"
+#include "printf.h"
 #include "psl/psl-phase.h"
 #include "tree.h"
 #include "type.h"
@@ -118,9 +119,9 @@ static void dump_type_hint(tree_t t)
 #if DUMP_TYPE_HINT
    static int nested = 0;
    if (++nested == 1 && tree_has_type(t)) {
-      color_printf("$red$$<$/*");
+      nvc_printf("$red$/*");
       dump_type(tree_type(t));
-      color_printf("$>$$red$*/$$");
+      nvc_printf("$red$*/$$");
    }
    --nested;
 #endif
@@ -987,7 +988,14 @@ static void dump_decl(tree_t t, int indent)
          case T_FOR_GENERATE: kind = "For generate"; break;
          case T_BLOCK: kind = "Block"; break;
          }
-         print_syntax("-- %s %s\n", kind, istr(tree_ident2(t)));
+
+         ident_t dotted = tree_ident(t);
+         ident_t symbol = tree_ident2(t);
+
+         print_syntax("-- %s %s", kind, istr(dotted));
+         if (dotted != symbol)
+            print_syntax(" (clone of %s)", istr(symbol));
+         print_syntax("\n");
       }
       return;
 
@@ -1074,24 +1082,24 @@ static void dump_alternative(tree_t t, int indent)
 {
    tab(indent);
    print_syntax("#when ");
+
    if (tree_has_ident(t))
       print_syntax("%s: ", istr(tree_ident(t)));
-   for (unsigned i = 0; i < tree_assocs(t); i++) {
-      if (i > 0) print_syntax("| ");
-      tree_t a = tree_assoc(t, i);
-      switch (tree_subkind(a)) {
-      case A_NAMED:
+
+   const int nchoices = tree_choices(t);
+   for (int i = 0; i < nchoices; i++) {
+      if (i > 0) print_syntax(" | ");
+      tree_t a = tree_choice(t, i);
+      if (tree_has_name(a))
          dump_expr(tree_name(a));
-         break;
-      case A_OTHERS:
-         print_syntax("#others");
-         break;
-      case A_RANGE:
+      else if (tree_ranges(a) > 0)
          dump_range(tree_range(a, 0));
-         break;
-      }
+      else
+         print_syntax("#others");
    }
+
    print_syntax(" =>\n");
+
    if (tree_decls(t) > 0) {
       dump_decls(t, indent + 4);
       tab(indent + 2);
@@ -1453,7 +1461,7 @@ static void dump_stmt(tree_t t, int indent)
    case T_SELECT:
       print_syntax(" <= ");
       if (tree_has_guard(t)) print_syntax("#guarded ");
-      color_printf("$red$/* TODO: T_SELECT */$$");
+      nvc_printf("$red$/* TODO: T_SELECT */$$");
       break;
 
    case T_CONCURRENT:
