@@ -214,7 +214,6 @@ static void pack_value(pack_writer_t *pw, jit_t *j, jit_value_t value)
       pack_uint(pw, value.int64);
       break;
    case JIT_ADDR_ABS:
-   case JIT_ADDR_COVER:
       pack_uint(pw, value.int64);
       break;
    case JIT_VALUE_LABEL:
@@ -497,7 +496,6 @@ static jit_value_t unpack_value(pack_func_t *pf, jit_t *j)
       value.int64 = unpack_uint(pf);
       break;
    case JIT_ADDR_ABS:
-   case JIT_ADDR_COVER:
       value.int64 = unpack_uint(pf);
       break;
    case JIT_VALUE_LABEL:
@@ -647,6 +645,7 @@ void jit_write_pack(jit_t *j, const ident_t *units, size_t count, FILE *f)
       write_fully(bytes, cpool_nbytes, f);
 
       write_fully(buf, size, f);
+      free(buf);
 
       if (func->cpoolsz > 0)
          write_fully(func->cpool, func->cpoolsz, f);
