@@ -66,10 +66,16 @@ static void mir_unit_free_memory(mir_unit_t *mu)
    ACLEAR(mu->linkage);
    ACLEAR(mu->extvars);
 
+   if (mu->shape != NULL) {
+      hash_free(mu->shape->objmap);
+      mu->shape->objmap = NULL;
+   }
+
    pool_free(mu->pool);
 
    hash_free(mu->objmap);
    hash_free(mu->privmap);
+   free(mu->vregs);
    free(mu->nodes);
    free(mu->argspill);
    free(mu);
@@ -405,3 +411,23 @@ ident_t mir_get_linkage(mir_unit_t *mu, unsigned nth)
 {
    return AGET(mu->linkage, nth);
 }
+
+#ifdef DEBUG
+void mir_compare_layout(mir_unit_t *a, mir_unit_t *b)
+{
+   if (a->vars.count != b->vars.count)
+      goto differ;
+
+   for (int i = 0; i < a->vars.count; i++) {
+      if (!mir_equals(a->vars.items[i].type, b->vars.items[i].type))
+         goto differ;
+   }
+
+   return;
+
+ differ:
+   mir_dump(a);
+   mir_dump(b);
+   fatal_trace("%s and %s have differing layout", istr(a->name), istr(b->name));
+}
+#endif
