@@ -1,5 +1,5 @@
 //
-//  Copyright (C) 2011-2025  Nick Gasson
+//  Copyright (C) 2011-2026  Nick Gasson
 //
 //  This program is free software: you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -20,13 +20,13 @@
 #include "cov/cov-api.h"
 #include "diag.h"
 #include "ident.h"
-#include "jit/jit-llvm.h"
 #include "jit/jit.h"
 #include "lib.h"
 #include "lower.h"
 #include "mir/mir-unit.h"
 #include "option.h"
 #include "phase.h"
+#include "printf.h"
 #include "rt/assert.h"
 #include "rt/model.h"
 #include "rt/mspace.h"
@@ -36,7 +36,7 @@
 #include "scan.h"
 #include "server.h"
 #include "thread.h"
-#include "vhpi/vhpi-util.h"
+#include "vhpi/vhpi-model.h"
 #include "vlog/vlog-node.h"
 #include "vlog/vlog-phase.h"
 #include "vpi/vpi-model.h"
@@ -59,12 +59,6 @@
 #define GIT_SHA_ONLY(x)
 #endif
 
-#if !defined HAVE_LLVM || !defined SYSTEM_CC
-#define DEFAULT_JIT true
-#else
-#define DEFAULT_JIT true
-#endif
-
 typedef struct {
    jit_t           *jit;
    unit_registry_t *registry;
@@ -81,7 +75,7 @@ typedef struct {
 } cmd_state_t;
 
 const char copy_string[] =
-   "Copyright (C) 2011-2025  Nick Gasson\n"
+   "Copyright (C) 2011-2026  Nick Gasson\n"
    "This program comes with ABSOLUTELY NO WARRANTY. This is free software, "
    "and\nyou are welcome to redistribute it under certain conditions. See "
    "the GNU\nGeneral Public Licence for details.";
@@ -153,6 +147,20 @@ static void parse_pp_define(char *optarg)
    }
 }
 
+static bool parse_warn_option(char *optarg)
+{
+   if (strcmp(optarg, "error") == 0)
+      return true;
+   else {
+      diag_t *d = diag_new(DIAG_FATAL, NULL);
+      diag_printf(d, "invalid warning option '%s'", optarg);
+      diag_hint(d, NULL, "valid options are: error");
+      diag_emit(d);
+
+      fatal_exit(EXIT_FAILURE);
+   }
+}
+
 static void do_file_list(const char *file, jit_t *jit, unit_registry_t *ur,
                          mir_context_t *mc)
 {
@@ -206,6 +214,19 @@ static void do_file_list(const char *file, jit_t *jit, unit_registry_t *ur,
    fclose(f);
 }
 
+static void werror_diag_cb(diag_t *d, void *ctx)
+{
+   diag_level_t level = diag_level(d, NULL);
+   if (level != DIAG_WARN)
+      return;
+
+   diag_hint(d, NULL, "this warning is treated as an error due to "
+             "$bold$-Werror$$");
+
+   level = DIAG_ERROR;
+   diag_level(d, &level);
+}
+
 static int analyse(int argc, char **argv, cmd_state_t *state)
 {
    static struct option long_options[] = {
@@ -222,14 +243,16 @@ static int analyse(int argc, char **argv, cmd_state_t *state)
       { "single-unit",     no_argument,       0, 'u' },
       { "preserve-case",   no_argument,       0, 'p' },
       { "keywords",        required_argument, 0, 'k' },
+      { "relative",        required_argument, 0, 'r' },
+      { "warn",            required_argument, 0, 'W' },
       { 0, 0, 0, 0 }
    };
 
    const int next_cmd = scan_cmd(2, argc, argv);
    int c, index = 0, error_limit = 20;
    const char *file_list = NULL;
-   const char *spec = ":D:f:I:";
-   bool no_save = false;
+   const char *spec = ":D:f:I:W:";
+   bool no_save = false, werror = false;
 
    while ((c = getopt_long(next_cmd, argv, spec, long_options, &index)) != -1) {
       switch (c) {
@@ -290,6 +313,12 @@ static int analyse(int argc, char **argv, cmd_state_t *state)
                fatal("'%s' is not a valid Verilog version", optarg);
          }
          break;
+      case 'r':
+         opt_set_str(OPT_RELATIVE_PATH, optarg);
+         break;
+      case 'W':
+         werror = parse_warn_option(optarg);
+         break;
       default:
          should_not_reach_here();
       }
@@ -297,13 +326,19 @@ static int analyse(int argc, char **argv, cmd_state_t *state)
 
    set_error_limit(error_limit);
 
+   if (werror)
+      diag_add_hint_fn(werror_diag_cb, NULL);
+
    if (state->mir == NULL)
       state->mir = mir_context_new();
 
    if (state->registry == NULL)
       state->registry = unit_registry_new(state->mir);
 
-   jit_t *jit = jit_new(state->registry, state->mir, NULL);
+   jit_t *jit = jit_new(state->registry, state->mir);
+
+   if (state->vhpi != NULL)
+      vhpi_run_callbacks(vhpiCbStartOfAnalysis);
 
    if (file_list != NULL)
       do_file_list(file_list, jit, state->registry, state->mir);
@@ -317,6 +352,12 @@ static int analyse(int argc, char **argv, cmd_state_t *state)
          analyse_file(argv[i], jit, state->registry, state->mir);
    }
 
+   if (state->vhpi != NULL)
+      vhpi_run_callbacks(vhpiCbEndOfAnalysis);
+
+   if (werror)
+      diag_remove_hint_fn(werror_diag_cb);
+
    jit_free(jit);
    set_error_limit(0);
 
@@ -454,13 +495,13 @@ static int elaborate(int argc, char **argv, cmd_state_t *state)
       { "verbose",         no_argument,       0, 'V' },
       { "no-save",         no_argument,       0, 'N' },
       { "jit",             no_argument,       0, 'j' },
-      { "precompile",      no_argument,       0, 'p' },   // DEPRECATED 1.18
       { "no-collapse",     no_argument,       0, 'C' },
+      { "stats",           no_argument,       0, 'S' },
       { "trace",           no_argument,       0, 't' },
       { 0, 0, 0, 0 }
    };
 
-   bool use_jit = DEFAULT_JIT, no_save = false;
+   bool no_save = false;
    unit_meta_t meta = {};
    cover_mask_t cover_mask = 0;
    const char *cover_spec_file = NULL, *sdf_args = NULL;
@@ -498,12 +539,7 @@ static int elaborate(int argc, char **argv, cmd_state_t *state)
          opt_set_int(OPT_NO_COLLAPSE, 1);
          break;
       case 'j':
-         use_jit = true;
-         break;
-      case 'p':
-         warnf("the $bold$--precompile$$ option is deprecated and will be "
-               "removed in a future release");
-         use_jit = false;
+         // No effect
          break;
       case 'g':
          parse_generic(optarg);
@@ -520,6 +556,9 @@ static int elaborate(int argc, char **argv, cmd_state_t *state)
       case 't':
          opt_set_int(OPT_RT_TRACE, 1);
          break;
+      case 'S':
+         opt_set_int(OPT_ELAB_STATS, 1);
+         break;
       case 0:
          // Set a flag
          break;
@@ -589,6 +628,8 @@ static int elaborate(int argc, char **argv, cmd_state_t *state)
    if (state->vhpi == NULL)
       state->vhpi = vhpi_context_new();
 
+   vhpi_run_callbacks(vhpiCbStartOfElaboration);
+
    tree_t top = elab(obj, state->jit, state->registry, state->mir,
                      state->cover, NULL, state->model);
 
@@ -599,45 +640,35 @@ static int elaborate(int argc, char **argv, cmd_state_t *state)
 
    progress("elaborating design");
 
-   if (state->cover != NULL) {
-      fbuf_t *f = fbuf_open(meta.cover_file, FBUF_OUT, FBUF_CS_NONE);
-      if (f == NULL)
-         fatal_errno("failed to open coverage database: %s", meta.cover_file);
-
-      cover_dump_items(state->cover, f, COV_DUMP_ELAB);
-      fbuf_close(f, NULL);
-
-      progress("dumping coverage data");
-   }
+   vhpi_run_callbacks(vhpiCbEndOfElaboration);
 
    if (error_count() > 0)
       return EXIT_FAILURE;
 
    const char *elab_name = istr(tree_ident(top));
    char *pack_name LOCAL = xasprintf("_%s.pack", elab_name);
-   char *dll_name LOCAL = xasprintf("_%s." DLL_EXT, elab_name);
 
    // Delete any existing generated code to avoid accidentally loading
    // the wrong version later
    lib_delete(state->work, pack_name);
-   lib_delete(state->work, dll_name);
 
    if (!no_save) {
       lib_save(state->work);
       progress("saving library");
    }
 
-   if (!use_jit)
-      cgen(top, state->registry, state->mir, state->jit, CGEN_NATIVE);
-   else if (!no_save)
-      cgen(top, state->registry, state->mir, state->jit, CGEN_JIT_PACK);
+   if (!no_save)
+      cgen(top, state->registry, state->mir, state->jit);
 
-   if (!use_jit || state->cover != NULL) {
-      // Must discard current JIT state to load AOT library later
-      model_free(state->model);
-      jit_free(state->jit);
-      state->jit = NULL;
-      state->model = NULL;
+   if (state->cover != NULL) {
+      fbuf_t *f = fbuf_open(meta.cover_file, FBUF_OUT, FBUF_CS_NONE);
+      if (f == NULL)
+         fatal_errno("failed to open coverage database: %s", meta.cover_file);
+
+      cover_write(state->cover, f, COV_DUMP_ELAB);
+      fbuf_close(f, NULL);
+
+      progress("dumping coverage data");
    }
 
    argc -= next_cmd - 1;
@@ -709,6 +740,17 @@ static void parse_exit_severity(const char *str)
    set_status_severity(s);
 }
 
+static void parse_stderr_severity(const char *str)
+{
+   diag_level_t level;
+   if (strcasecmp(str, "none") == 0)
+      level = DIAG_FATAL + 1;  // No messages to stderr
+   else
+      level = get_diag_severity(parse_severity(optarg));
+
+   opt_set_int(OPT_STDERR_LEVEL, level);
+}
+
 static int parse_stop_delta(const char *str)
 {
    const int ival = parse_int(str);
@@ -731,7 +773,7 @@ static void ctrl_c_handler(void *arg)
 
 static jit_t *get_jit(cmd_state_t *state)
 {
-   jit_t *jit = jit_new(state->registry, state->mir, state->cover);
+   jit_t *jit = jit_new(state->registry, state->mir);
 
 #if defined HAVE_LLVM && 1
    jit_register_llvm_plugin(jit);
@@ -770,7 +812,7 @@ static cover_data_t *load_coverage(const unit_meta_t *meta)
    if (f == NULL)
       fatal_errno("failed to open coverage database: %s", meta->cover_file);
 
-   cover_data_t *db = cover_read_items(f, 0);
+   cover_data_t *db = cover_read(f, 0);
 
    fbuf_close(f, NULL);
    return db;
@@ -784,7 +826,7 @@ static void emit_coverage(const unit_meta_t *meta, jit_t *j, cover_data_t *db)
    if (f == NULL)
       fatal_errno("failed to open coverage database: %s", meta->cover_file);
 
-   cover_dump_items(db, f, COV_DUMP_RUNTIME);
+   cover_write(db, f, COV_DUMP_RUNTIME);
 
    fbuf_close(f, NULL);
 }
@@ -1001,10 +1043,6 @@ static int run_cmd(int argc, char **argv, cmd_state_t *state)
    if (state->jit == NULL)
       state->jit = get_jit(state);
 
-#ifdef ENABLE_LLVM
-   jit_load_dll(state->jit, tree_ident(top));
-#endif
-
    load_jit_pack(state->jit, top);
 
    if (state->model == NULL) {
@@ -1018,9 +1056,11 @@ static int run_cmd(int argc, char **argv, cmd_state_t *state)
    if (state->vpi == NULL)
       state->vpi = vpi_context_new();
 
+   if (nplusargs > 0)
+      vhpi_set_plusargs(state->vhpi, nplusargs, plusargs);
+
    if (pli_plugins != NULL || state->plugins != NULL) {
-      vhpi_context_initialise(state->vhpi, top, state->model, state->jit,
-                              nplusargs, plusargs);
+      vhpi_context_initialise(state->vhpi, top, state->model, state->jit);
       vpi_context_initialise(state->vpi, top, state->model, state->jit,
                              nplusargs, plusargs);
    }
@@ -1495,10 +1535,6 @@ static int do_cmd(int argc, char **argv, cmd_state_t *state)
       if (top == NULL)
          fatal("%s not elaborated", istr(state->top_level));
 
-#ifdef ENABLE_LLVM
-      jit_load_dll(state->jit, tree_ident(top));
-#endif
-
       load_jit_pack(state->jit, top);
 
       shell_reset(sh, top);
@@ -1573,10 +1609,6 @@ static int interact_cmd(int argc, char **argv, cmd_state_t *state)
       if (top == NULL)
          fatal("%s not elaborated", istr(state->top_level));
 
-#ifdef ENABLE_LLVM
-      jit_load_dll(state->jit, tree_ident(top));
-#endif
-
       load_jit_pack(state->jit, top);
 
       shell_reset(sh, top);
@@ -1686,7 +1718,7 @@ static int coverage_cmd(int argc, char **argv, cmd_state_t *state)
       case ':':
          missing_argument("coverage", argv);
       default:
-         abort();
+         should_not_reach_here();
       }
    }
 
@@ -1695,7 +1727,7 @@ static int coverage_cmd(int argc, char **argv, cmd_state_t *state)
    if (optind == argc)
       fatal("no input coverage database FILE specified");
 
-   cover_data_t *cover = NULL;
+   cover_data_t *merged = NULL;
 
    // Rest of inputs are coverage input files
    for (int i = optind; i < argc; i++) {
@@ -1703,10 +1735,14 @@ static int coverage_cmd(int argc, char **argv, cmd_state_t *state)
 
       if (f != NULL) {
          progress("Loading input coverage database: %s", argv[i]);
+
+         cover_data_t *db = cover_read(f, rpt_mask);
          if (i == optind)
-            cover = cover_read_items(f, rpt_mask);
-         else
-            cover_merge_items(f, cover, MERGE_UNION);
+            merged = db;
+         else {
+            cover_merge(merged, db, MERGE_UNION);
+            // TODO: not safe to free db here
+         }
       }
       else
          fatal("Could not open coverage database: %s", argv[i]);
@@ -1717,25 +1753,25 @@ static int coverage_cmd(int argc, char **argv, cmd_state_t *state)
    if (out_db) {
       progress("Saving merged coverage database to: %s", out_db);
       fbuf_t *f = fbuf_open(out_db, FBUF_OUT, FBUF_CS_NONE);
-      cover_dump_items(cover, f, COV_DUMP_PROCESSING);
+      cover_write(merged, f, COV_DUMP_PROCESSING);
       fbuf_close(f, NULL);
    }
 
-   if (exclude_file && cover) {
+   if (exclude_file && merged) {
       progress("Loading exclude file: %s", exclude_file);
-      cover_load_exclude_file(exclude_file, cover);
+      cover_load_exclude_file(exclude_file, merged);
    }
 
-   if (rpt_file && cover) {
+   if (rpt_file && merged) {
       progress("Generating code coverage report.");
-      cover_report(rpt_file, cover, item_limit);
+      cover_report(rpt_file, merged, item_limit);
    }
 
-   if (export_file && cover) {
+   if (export_file && merged) {
       progress("Exporting XML coverage report");
 
       FILE *f = create_file("%s", export_file);
-      cover_export_cobertura(cover, f, NULL);
+      cover_export_cobertura(merged, f, NULL);
       fclose(f);
    }
 
@@ -1820,7 +1856,7 @@ static cover_data_t *merge_coverage_files(int argc, int next_cmd, char **argv,
    if (optind == next_cmd)
       fatal("no input coverage database specified");
 
-   cover_data_t *cover = NULL;
+   cover_data_t *merged = NULL;
 
    for (int i = optind; i < next_cmd; i++) {
       fbuf_t *f = fbuf_open(argv[i], FBUF_IN, FBUF_CS_NONE);
@@ -1851,15 +1887,17 @@ static cover_data_t *merge_coverage_files(int argc, int next_cmd, char **argv,
 
       progress("loading input coverage database %s", argv[i]);
 
+      cover_data_t *db = cover_read(f, rpt_mask);
+
       if (i == optind)
-         cover = cover_read_items(f, rpt_mask);
+         merged = db;
       else
-         cover_merge_items(f, cover, mode);
+         cover_merge(merged, db, mode);
 
       fbuf_close(f, NULL);
    }
 
-   return cover;
+   return merged;
 }
 
 static int cover_export_cmd(int argc, char **argv, cmd_state_t *state)
@@ -1934,7 +1972,7 @@ static int cover_export_cmd(int argc, char **argv, cmd_state_t *state)
       if (f == NULL)
          fatal("no coverage database for %s", istr(state->top_level));
 
-      cover = cover_read_items(f, 0);
+      cover = cover_read(f, 0);
       fbuf_close(f, NULL);
 
       warnf("exporting the coverage database using the top-level unit name "
@@ -2087,7 +2125,7 @@ static int cover_merge_cmd(int argc, char **argv, cmd_state_t *state)
    progress("saving merged coverage database to %s", out_db);
 
    fbuf_t *f = fbuf_open(out_db, FBUF_OUT, FBUF_CS_NONE);
-   cover_dump_items(cover, f, COV_DUMP_PROCESSING);
+   cover_write(cover, f, COV_DUMP_PROCESSING);
    fbuf_close(f, NULL);
 
    argc -= next_cmd - 1;
@@ -2148,8 +2186,8 @@ static int preprocess_cmd(int argc, char **argv, cmd_state_t *state)
 
 static void usage(void)
 {
-   color_printf("$!cyan$Usage:$$ $bold$%s [OPTION]... "
-                "COMMAND [OPTION]...$$\n\n",  PACKAGE);
+   nvc_printf("$!cyan$Usage:$$ $bold$%s [OPTION]... "
+              "COMMAND [OPTION]...$$\n\n",  PACKAGE);
 
    wrapped_printf("Global options are placed before COMMAND, and "
                   "command-specific options are placed afterwards. "
@@ -2206,7 +2244,7 @@ static void usage(void)
              "Diagnostic message style, compact is less verbose" },
            { "--seed=N", "Seed for random number generation" },
            { "--std={1993,..,2019}", "VHDL standard revision to use" },
-           { "--stderr={note,warning,error,failure}",
+           { "--stderr={note,warning,error,failure,none}",
              "Print messages of this severity level or higher to stderr" },
            { "-v, --version", "Display version and copyright information" },
            { "--vhpi-debug", "Report VHPI errors as diagnostic messages" },
@@ -2229,9 +2267,11 @@ static void usage(void)
            { "--preserve-case",
              "Preserve the original case of VHDL identifiers" },
            { "--psl", "Enable parsing of PSL directives in comments" },
+           { "--relative=DIR", "Store paths relative to DIR" },
            { "--relaxed", "Disable certain pedantic rule checks" },
            { "--single-unit",
              "Treat all Verilog files as a single compilation unit" },
+           { "-Werror", "Treat all analysis warnings as errors" },
         }
       },
       { "Elaboration options",
@@ -2248,6 +2288,7 @@ static void usage(void)
            { "-O0, -O1, -O2, -O3", "Set optimisation level (default is -O2)" },
            { "--no-collapse", "Do not collapse multiple signals into one" },
            { "--no-save", "Do not save the elaborated design to disk" },
+           { "--stats", "Print statistics about instantiated design units" },
            { "-V, --verbose", "Print resource usage at each step" },
         }
       },
@@ -2305,7 +2346,7 @@ static void usage(void)
         {
            { "--format=FMT", "File format (must be 'cobertura')" },
            { "-o, --output=FILE", "Output file name" },
-           { "--relative=PATH", "Strip PATH from prefix of absolute paths" },
+           { "--relative=PATH", "Report file names relative to PATH" },
         }
       },
       { "Install options",
@@ -2318,7 +2359,7 @@ static void usage(void)
    const int right = MAX(60, terminal_width());
 
    for (int i = 0; i < ARRAY_LEN(groups); i++) {
-      color_printf("$bold$$cyan$%s:$$\n", groups[i].group);
+      nvc_printf("$bold$$cyan$%s:$$\n", groups[i].group);
 
       for (int j = 0; j < ARRAY_LEN(groups[i].options); j++) {
          const char *args  = groups[i].options[j].args;
@@ -2329,9 +2370,9 @@ static void usage(void)
 
          int col = 0;
          if (args[0] == '-' && args[1] == '-' && i > 0)
-            col += color_printf("     $bold$%s$$ ", args);
+            col += nvc_printf("     $bold$%s$$ ", args);
          else
-            col += color_printf(" $bold$%s$$ ", args);
+            col += nvc_printf(" $bold$%s$$ ", args);
 
          const int indent = i == 0 ? 30 : 20;
          if (col > indent)
@@ -2367,13 +2408,13 @@ static void usage(void)
 
    LOCAL_TEXT_BUF tb = tb_new();
    lib_print_search_paths(tb);
-   color_printf("$!cyan$Library search paths:$$%s\n\n", tb_get(tb));
+   nvc_printf("$!cyan$Library search paths:$$%s\n\n", tb_get(tb));
 
    wrapped_printf("The full manual can be read with $bold$man 1 %s$$ and "
                   "contains detailed explanations of the commands and options "
                   "above as well as examples.\n", PACKAGE_NAME);
-   color_printf("\nReport bugs at $link:%s\07%s$\n", PACKAGE_BUGREPORT,
-                PACKAGE_BUGREPORT);
+   nvc_printf("\nReport bugs at $link:%s\07%s$\n", PACKAGE_BUGREPORT,
+              PACKAGE_BUGREPORT);
 }
 
 static vhdl_standard_t parse_standard(const char *str)
@@ -2610,7 +2651,7 @@ int main(int argc, char **argv)
          opt_set_size(OPT_HEAP_SIZE, parse_size(optarg));
          break;
       case 'E':
-         set_stderr_severity(parse_severity(optarg));
+         parse_stderr_severity(optarg);
          break;
       case 'l':
          state.plugins = optarg;
@@ -2645,10 +2686,25 @@ int main(int argc, char **argv)
    argc -= next_cmd - 1;
    argv += next_cmd - 1;
 
+   // Shuffle the arguments to put all the plusargs first
+   qsort(argv + optind, next_cmd - optind, sizeof(char *), plusarg_cmp);
+
+   int nplusargs = 0;
+   char **plusargs = argv + optind;
+   for (int i = optind; i < next_cmd; i++) {
+      if (argv[i][0] == '+')
+         nplusargs++, optind++;
+   }
+
    if (state.plugins != NULL) {
       state.vhpi = vhpi_context_new();
       vhpi_load_plugins(state.plugins);
+
+      if (nplusargs > 0)
+         vhpi_set_plusargs(state.vhpi, nplusargs, plusargs);
    }
+   else if (nplusargs > 0)
+      warnf("found plusargs on command line but no VHPI plugin was loaded");
 
    const int ret = process_command(argc, argv, &state);
 
