@@ -56,7 +56,6 @@ typedef struct _object_arena {
    vhdl_standard_t std;
    uint32_t        checksum;
    generation_t    copygen;
-   bool            copyflag;
    bool            frozen;
    bool            obsolete;
 } object_arena_t;
@@ -77,7 +76,7 @@ typedef struct _object_arena {
                           | I_GENMAPS | I_ASSOCS | I_CONTEXT            \
                           | I_LITERALS | I_FIELDS | I_UNITS | I_CHARS   \
                           | I_DIMS | I_RANGES | I_INDEXES | I_PARTS \
-                          | I_PRAGMAS)
+                          | I_PRAGMAS | I_CHOICES)
 #define ITEM_INT64       (I_POS | I_IVAL)
 #define ITEM_INT32       (I_SUBKIND | I_CLASS | I_FLAGS)
 #define ITEM_DOUBLE      (I_DVAL)
@@ -91,7 +90,7 @@ static const char *item_text_map[] = {
    "I_REF",      "I_FILE_MODE",  "I_ASSOCS",   "I_CONTEXT",    "I_TRIGGERS",
    "I_PARTS"  ,  "I_CLASS",      "I_RANGES",   "I_NAME",       "I_PRAGMAS",
    "I_DVAL",     "I_SPEC",       "I_FOREIGN",  "I_INDEXES",    "I_BASE",
-   "I_ELEM",     "I_DESIGNATED", "???",        "I_RESOLUTION", "I_RESULT",
+   "I_ELEM",     "I_DESIGNATED", "I_CHOICES",  "I_RESOLUTION", "I_RESULT",
    "I_UNITS",    "I_LITERALS",   "I_DIMS",     "I_FIELDS",     "I_CLOCK",
    "I_GUARD",    "???",          "I_CHARS",    "I_CONSTRAINT", "I_FLAGS",
    "???",        "I_LEFT",       "I_RIGHT",    "I_NUMBER",     "I_MESSAGE",
@@ -119,15 +118,9 @@ static inline object_arena_t *__object_arena(object_t *object)
 static ident_t object_arena_name(object_arena_t *arena)
 {
    if (arena->alloc > arena->base) {
-      object_t *root = arena_root(arena);
-
-      const object_class_t *class = classes[root->tag];
-      const imask_t has = class->has_map[root->kind];
-
-      if (has & I_IDENT) {
-         const int n = __builtin_popcountll(has & (I_IDENT - 1));
-         return root->items[n].ident;
-      }
+      ident_t name = object_ident(arena_root(arena));
+      if (name != NULL)
+         return name;
    }
 
    return ident_new("???");
@@ -219,6 +212,19 @@ object_arena_t *object_arena(object_t *object)
    return __object_arena(object);
 }
 
+ident_t object_ident(object_t *object)
+{
+   const object_class_t *class = classes[object->tag];
+   const imask_t has = class->has_map[object->kind];
+
+   if (has & I_IDENT) {
+      const int n = __builtin_popcountll(has & (I_IDENT - 1));
+      return object->items[n].ident;
+   }
+
+   return NULL;
+}
+
 void __object_write_barrier(object_t *lhs, object_t *rhs)
 {
    const uintptr_t lhs_mask = (uintptr_t)lhs & ~OBJECT_PAGE_MASK;
@@ -278,6 +284,26 @@ void obj_array_add(obj_array_t **a, object_t *o)
    (*a)->items[(*a)->count++] = o;
 }
 
+void obj_array_copy(obj_array_t **dst, const obj_array_t *src)
+{
+   if (src == NULL)
+      return;
+
+   if (*dst == NULL) {
+      *dst = xmalloc_flex(sizeof(obj_array_t), src->count, sizeof(object_t *));
+      (*dst)->count = 0;
+      (*dst)->limit = src->count;
+   }
+   else if ((*dst)->count + src->count > (*dst)->limit) {
+      (*dst)->limit = (*dst)->count + src->count;
+      *dst = xrealloc_flex(*dst, sizeof(obj_array_t),
+                           (*dst)->limit, sizeof(object_t *));
+   }
+
+   for (int i = 0; i < src->count; i++)
+      (*dst)->items[(*dst)->count++] = src->items[i];
+}
+
 void obj_array_free(obj_array_t **a)
 {
    free(*a);
@@ -1037,7 +1063,7 @@ static bool object_copy_mark(object_t *object, object_copy_ctx_t *ctx)
       return false;
 
    object_arena_t *arena = __object_arena(object);
-   if (arena->copygen != ctx->generation || !arena->copyflag)
+   if (arena->copygen != ctx->generation)
       return false;
 
    if (ctx->copy_map == NULL)
@@ -1091,31 +1117,10 @@ static object_t *object_copy_map(object_t *object, object_copy_ctx_t *ctx)
    return map ?: object;
 }
 
-static bool object_copy_root_closure(object_arena_t *a, object_copy_ctx_t *ctx)
-{
-   bool include = false;
-
-   if (a->copygen == ctx->generation)
-      return a->copyflag;
-
-   for (int i = 0; i < ctx->nroots; i++)
-      include |= (__object_arena(ctx->roots[i]) == a);
-
-   for (int i = 0; i < a->deps.count; i++)
-      include |= object_copy_root_closure(a->deps.items[i], ctx);
-
-   a->copygen  = ctx->generation;
-   a->copyflag = include;
-
-   return include;
-}
-
 void object_copy(object_copy_ctx_t *ctx)
 {
-   for (int i = 0; i < ctx->nroots; i++) {
-      object_arena_t *a = __object_arena(ctx->roots[i]);
-      object_copy_root_closure(a, ctx);
-   }
+   for (int i = 0; i < ctx->nroots; i++)
+      __object_arena(ctx->roots[i])->copygen = ctx->generation;
 
    for (int i = 0; i < ctx->nroots; i++)
       (void)object_copy_mark(ctx->roots[i], ctx);
@@ -1252,7 +1257,7 @@ void object_arena_freeze(object_arena_t *arena)
 void arena_walk_deps(object_arena_t *arena, arena_deps_fn_t fn, void *context)
 {
    for (unsigned i = 0; i < arena->deps.count; i++)
-      (*fn)(object_arena_name(arena->deps.items[i]), context);
+      (*fn)(arena_root(arena->deps.items[i]), context);
 }
 
 void arena_walk_obsolete_deps(object_arena_t *arena, arena_deps_fn_t fn,
@@ -1260,7 +1265,7 @@ void arena_walk_obsolete_deps(object_arena_t *arena, arena_deps_fn_t fn,
 {
    for (unsigned i = 0; i < arena->deps.count; i++) {
       if (arena->deps.items[i]->obsolete)
-         (*fn)(object_arena_name(arena->deps.items[i]), context);
+         (*fn)(arena_root(arena->deps.items[i]), context);
    }
 }
 
@@ -1355,3 +1360,18 @@ void make_new_arena(void)
    freeze_global_arena();
    global_arena = object_arena_new(object_arena_default_size(), standard());
 }
+
+void discard_global_arena(void)
+{
+   object_one_time_init();
+
+   if (global_arena != NULL) {
+      nvc_munmap(global_arena->base, global_arena->limit - global_arena->base);
+
+      assert(all_arenas.items[all_arenas.count - 1] == global_arena);
+      APOP(all_arenas);
+
+      free(global_arena);
+      global_arena = NULL;
+   }
+}
