@@ -71,7 +71,7 @@ typedef uint64_t imask_t;
 #define I_BASE       ONE_HOT(34)
 #define I_ELEM       ONE_HOT(35)
 #define I_DESIGNATED ONE_HOT(36)
-// Unused            ONE_HOT(37)
+#define I_CHOICES    ONE_HOT(37)
 #define I_RESOLUTION ONE_HOT(38)
 #define I_RESULT     ONE_HOT(39)
 #define I_UNITS      ONE_HOT(40)
@@ -142,6 +142,7 @@ typedef struct {
       })
 
 void obj_array_add(obj_array_t **a, object_t *o);
+void obj_array_copy(obj_array_t **dst, const obj_array_t *src);
 
 typedef union {
    ident_t       ident;
@@ -235,6 +236,7 @@ object_t *object_rewrite(object_t *object, object_rewrite_ctx_t *ctx);
 unsigned object_next_generation(void);
 void object_copy(object_copy_ctx_t *ctx);
 object_arena_t *object_arena(object_t *object);
+ident_t object_ident(object_t *object);
 size_t object_arena_default_size(void);
 object_t *arena_root(object_arena_t *arena);
 void arena_set_checksum(object_arena_t *arena, uint32_t checksum);
@@ -260,7 +262,7 @@ void __object_write_barrier(object_t *lhs, object_t *rhs);
 object_arena_t *object_arena_new(size_t size, unsigned std);
 void object_arena_freeze(object_arena_t *arena);
 
-typedef void (*arena_deps_fn_t)(ident_t, void *);
+typedef void (*arena_deps_fn_t)(object_t *, void *);
 void arena_walk_deps(object_arena_t *arena, arena_deps_fn_t fn, void *context);
 void arena_walk_obsolete_deps(object_arena_t *arena, arena_deps_fn_t fn,
                               void *context);
@@ -271,5 +273,6 @@ object_t *object_from_locus(ident_t module, ptrdiff_t offset,
 
 void make_new_arena(void);
 void freeze_global_arena(void);
+void discard_global_arena(void);
 
 #endif   // _OBJECT_H
