@@ -328,7 +328,7 @@ static const imask_t has_map[T_LAST_TREE_KIND] = {
    (I_IDENT | I_VALUE | I_STMTS),
 
    // T_ALTERNATIVE
-   (I_IDENT | I_ASSOCS | I_STMTS | I_DECLS),
+   (I_IDENT | I_CHOICES | I_STMTS | I_DECLS),
 
    // T_PSL_DECL
    (I_IDENT | I_FOREIGN),
@@ -367,7 +367,7 @@ static const imask_t has_map[T_LAST_TREE_KIND] = {
    (I_REF | I_SPEC | I_TYPE),
 
    // T_INERTIAL
-   (I_VALUE | I_TYPE | I_TARGET),
+   (I_VALUE | I_TYPE),
 
    // T_ELEM_RESOLUTION
    (I_ASSOCS),
@@ -385,7 +385,10 @@ static const imask_t has_map[T_LAST_TREE_KIND] = {
    (I_FOREIGN | I_TYPE),
 
    // T_PSL_UNION
-   (I_FOREIGN | I_TYPE)
+   (I_FOREIGN | I_TYPE),
+
+   // T_CHOICE
+   (I_NAME | I_RANGES),
 };
 
 static const char *kind_text_map[T_LAST_TREE_KIND] = {
@@ -428,7 +431,7 @@ static const char *kind_text_map[T_LAST_TREE_KIND] = {
    "T_MATCH_SELECT",    "T_PROT_DECL",       "T_DUMMY_DRIVER",
    "T_GUARD",           "T_INERTIAL",        "T_ELEM_RESOLUTION",
    "T_LOOP",            "T_REPORT",          "T_PSL_DIRECT",
-   "T_PSL_FCALL",       "T_PSL_UNION",
+   "T_PSL_FCALL",       "T_PSL_UNION",       "T_CHOICE",
 };
 
 static const change_allowed_t change_allowed[] = {
@@ -481,6 +484,11 @@ object_class_t tree_object = {
    .gc_num_roots   = 9
 };
 
+typedef struct {
+   tree_deps_fn_t fn;
+   void          *context;
+} tree_deps_args_t;
+
 #ifdef DEBUG
 static tree_kind_t expr_kinds[] = {
    T_FCALL,     T_LITERAL,       T_REF,        T_QUALIFIED,
@@ -538,6 +546,19 @@ static inline void tree_array_add(item_t *item, tree_t t)
    obj_array_add(&(item->obj_array), &(t->object));
 }
 
+static inline void tree_copy_array(tree_t t, tree_t from, imask_t mask)
+{
+   item_t *dst = lookup_item(&tree_object, t, mask);
+   const item_t *src = lookup_item(&tree_object, from, mask);
+   if (src->obj_array == NULL)
+      return;
+
+   obj_array_copy(&(dst->obj_array), src->obj_array);
+
+   for (int i = 0; i < src->obj_array->count; i++)
+      object_write_barrier(&(t->object), src->obj_array->items[i]);
+}
+
 tree_t tree_new(tree_kind_t kind)
 {
    object_t *o = object_new(NULL, &tree_object, kind);
@@ -622,6 +643,11 @@ void tree_add_port(tree_t t, tree_t d)
    object_write_barrier(&(t->object), &(d->object));
 }
 
+void tree_copy_ports(tree_t t, tree_t from)
+{
+   tree_copy_array(t, from, I_PORTS);
+}
+
 unsigned tree_subkind(tree_t t)
 {
    item_t *item = lookup_item(&tree_object, t, I_SUBKIND);
@@ -652,6 +678,11 @@ void tree_add_generic(tree_t t, tree_t d)
    object_write_barrier(&(t->object), &(d->object));
 }
 
+void tree_copy_generics(tree_t t, tree_t from)
+{
+   tree_copy_array(t, from, I_GENERICS);
+}
+
 type_t tree_type(tree_t t)
 {
    item_t *item = lookup_item(&tree_object, t, I_TYPE);
@@ -894,6 +925,11 @@ void tree_add_decl(tree_t t, tree_t d)
    object_write_barrier(&(t->object), &(d->object));
 }
 
+void tree_copy_decls(tree_t t, tree_t from)
+{
+   tree_copy_array(t, from, I_DECLS);
+}
+
 unsigned tree_stmts(tree_t t)
 {
    item_t *item = lookup_item(&tree_object, t, I_STMTS);
@@ -913,6 +949,11 @@ void tree_add_stmt(tree_t t, tree_t s)
    object_write_barrier(&(t->object), &(s->object));
 }
 
+void tree_copy_stmts(tree_t t, tree_t from)
+{
+   tree_copy_array(t, from, I_STMTS);
+}
+
 unsigned tree_waveforms(tree_t t)
 {
    item_t *item = lookup_item(&tree_object, t, I_WAVES);
@@ -1098,6 +1139,25 @@ void tree_add_assoc(tree_t t, tree_t a)
    object_write_barrier(&(t->object), &(a->object));
 }
 
+unsigned tree_choices(tree_t t)
+{
+   item_t *item = lookup_item(&tree_object, t, I_CHOICES);
+   return obj_array_count(item->obj_array);
+}
+
+tree_t tree_choice(tree_t t, unsigned n)
+{
+   item_t *item = lookup_item(&tree_object, t, I_CHOICES);
+   return tree_array_nth(item, n);
+}
+
+void tree_add_choice(tree_t t, tree_t c)
+{
+   assert(c->object.kind == T_CHOICE);
+   tree_array_add(lookup_item(&tree_object, t, I_CHOICES), c);
+   object_write_barrier(&(t->object), &(c->object));
+}
+
 tree_t tree_severity(tree_t t)
 {
    item_t *item = lookup_item(&tree_object, t, I_SEVERITY);
@@ -1375,9 +1435,18 @@ tree_t tree_container(tree_t t)
    return container_of(o, struct _tree, object);
 }
 
+static void tree_deps_cb(object_t *obj, void *ctx)
+{
+   tree_deps_args_t *args = ctx;
+   tree_t t = tree_from_object(obj);
+   if (t != NULL)
+      (*args->fn)(t, args->context);
+}
+
 void tree_walk_deps(tree_t t, tree_deps_fn_t fn, void *ctx)
 {
-   arena_walk_deps(object_arena(&(t->object)), fn, ctx);
+   tree_deps_args_t args = { fn, ctx };
+   arena_walk_deps(object_arena(&(t->object)), tree_deps_cb, &args);
 }
 
 object_t *tree_to_object(tree_t t)
