@@ -208,6 +208,14 @@ typedef enum {
    S_TO_STRING_REAL_FORMAT,
    S_TO_HSTRING_BITVEC,
    S_TO_OSTRING_BITVEC,
+   S_IEEE_MISC,
+   S_IEEE_AND,
+   S_IEEE_OR,
+   S_IEEE_XOR,
+   S_IEEE_NAND,
+   S_IEEE_NOR,
+   S_IEEE_XNOR,
+   S_IEEE_NOT,
 } subprogram_kind_t;
 
 typedef enum {
@@ -394,6 +402,7 @@ typedef enum tree_kind {
    T_PSL_DIRECT,
    T_PSL_FCALL,
    T_PSL_UNION,
+   T_CHOICE,
 
    T_LAST_TREE_KIND
 } tree_kind_t;
@@ -459,6 +468,7 @@ bool tree_has_ident2(tree_t t);
 unsigned tree_ports(tree_t t);
 tree_t tree_port(tree_t t, unsigned n);
 void tree_add_port(tree_t t, tree_t d);
+void tree_copy_ports(tree_t t, tree_t from);
 
 tree_t tree_file_mode(tree_t t);
 void tree_set_file_mode(tree_t t, tree_t m);
@@ -466,6 +476,7 @@ void tree_set_file_mode(tree_t t, tree_t m);
 unsigned tree_generics(tree_t t);
 tree_t tree_generic(tree_t t, unsigned n);
 void tree_add_generic(tree_t t, tree_t d);
+void tree_copy_generics(tree_t t, tree_t from);
 
 unsigned tree_genmaps(tree_t t);
 tree_t tree_genmap(tree_t t, unsigned n);
@@ -497,10 +508,12 @@ void tree_add_waveform(tree_t t, tree_t w);
 unsigned tree_decls(tree_t t);
 tree_t tree_decl(tree_t t, unsigned n);
 void tree_add_decl(tree_t t, tree_t d);
+void tree_copy_decls(tree_t t, tree_t from);
 
 unsigned tree_stmts(tree_t t);
 tree_t tree_stmt(tree_t t, unsigned n);
-void tree_add_stmt(tree_t t, tree_t d);
+void tree_add_stmt(tree_t t, tree_t s);
+void tree_copy_stmts(tree_t t, tree_t from);
 
 unsigned tree_conds(tree_t t);
 tree_t tree_cond(tree_t t, unsigned n);
@@ -533,6 +546,10 @@ unsigned tree_assocs(tree_t t);
 tree_t tree_assoc(tree_t t, unsigned n);
 void tree_add_assoc(tree_t t, tree_t a);
 
+unsigned tree_choices(tree_t t);
+tree_t tree_choice(tree_t t, unsigned n);
+void tree_add_choice(tree_t t, tree_t c);
+
 tree_t tree_severity(tree_t t);
 bool tree_has_severity(tree_t t);
 void tree_set_severity(tree_t t, tree_t s);
@@ -628,7 +645,7 @@ void tree_copy(tree_t *roots, unsigned nroots,
                void *callback_context);
 
 typedef tree_t (*tree_load_fn_t)(ident_t);
-typedef void (*tree_deps_fn_t)(ident_t, void *);
+typedef void (*tree_deps_fn_t)(tree_t, void *);
 
 object_arena_t *tree_arena(tree_t t);
 tree_t tree_container(tree_t t);
