@@ -1,5 +1,5 @@
 //
-//  Copyright (C) 2025 Nick Gasson
+//  Copyright (C) 2025-2026 Nick Gasson
 //
 //  This program is free software: you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -20,6 +20,10 @@
 
 #include "prim.h"
 
-tree_t vhdl_generate_instance(tree_t t, ident_t prefix, ident_t ndotted);
+tree_t vhdl_component_instance(tree_t comp, tree_t inst, ident_t dotted);
+tree_t vhdl_architecture_instance(tree_t arch, tree_t inst, ident_t dotted);
+tree_t vhdl_config_instance(tree_t conf, tree_t bind, ident_t dotted);
+
+void vhdl_cover_block(tree_t block, cover_data_t *db, cover_scope_t *cs);
 
 #endif  // _VHDL_PHASE_H
