@@ -122,7 +122,7 @@ static const imask_t has_map[V_LAST_NODE_KIND] = {
    (I_PARAMS),
 
    // V_DATA_TYPE
-   (I_SUBKIND | I_RANGES),
+   (I_SUBKIND | I_RANGES | I_FLAGS),
 
    // V_TYPE_DECL
    (I_IDENT | I_TYPE),
@@ -266,7 +266,7 @@ static const imask_t has_map[V_LAST_NODE_KIND] = {
    (I_IDENT | I_IDENT2 | I_REF),
 
    // V_TF_PORT_DECL
-   (I_IDENT | I_SUBKIND | I_RANGES | I_TYPE | I_VALUE),
+   (I_IDENT | I_SUBKIND | I_RANGES | I_TYPE | I_VALUE | I_REF),
 
    // V_RETURN
    (I_REF | I_VALUE),
@@ -274,14 +274,41 @@ static const imask_t has_map[V_LAST_NODE_KIND] = {
    // V_OP_ASSIGN
    (I_TARGET | I_SUBKIND | I_VALUE),
 
-   // V_STRUCT_REF
+   // V_MEMBER_REF
    (I_VALUE | I_IDENT | I_REF),
 
    // V_PACKAGE
-   (I_DECLS | I_IDENT),
+   (I_DECLS | I_IDENT | I_IDENT2),
 
    // V_MIN_TYP_MAX
-   (I_LEFT | I_VALUE | I_RIGHT)
+   (I_LEFT | I_VALUE | I_RIGHT),
+
+   // V_PROGRAM
+   (I_IDENT | I_STMTS | I_DECLS | I_IDENT2),
+
+   // V_CLASS_DECL
+   (I_IDENT | I_IDENT2 | I_DECLS),
+
+   // V_NULL
+   (I_TYPE),
+
+   // V_CLASS_NEW
+   (I_TYPE | I_PARAMS),
+
+   // V_CONSTRUCTOR
+   (I_IDENT | I_PARAMS | I_STMTS | I_DECLS),
+
+   // V_SUPER_CALL
+   (I_PARAMS),
+
+   // V_IMPORT_DECL
+   (I_IDENT | I_IDENT2),
+
+   // V_NAMESPACE
+   (I_DECLS),
+
+   // V_DEFPARAM
+   (I_TARGET | I_VALUE),
 };
 
 static const char *kind_text_map[V_LAST_NODE_KIND] = {
@@ -305,8 +332,10 @@ static const char *kind_text_map[V_LAST_NODE_KIND] = {
    "V_SPECPARAM",     "V_FORK",        "V_ATTR_INST",     "V_USER_TCALL",
    "V_VOID_CALL",     "V_GENVAR_DECL", "V_FOR_GENERATE",  "V_DEASSIGN",
    "V_FORCE",         "V_RELEASE",     "V_DISABLE",       "V_HIER_REF",
-   "V_TF_PORT_DECL",  "V_RETURN",      "V_OP_ASSIGN",     "V_STRUCT_REF",
-   "V_PACKAGE",       "V_MIN_TYP_MAX"
+   "V_TF_PORT_DECL",  "V_RETURN",      "V_OP_ASSIGN",     "V_MEMBER_REF",
+   "V_PACKAGE",       "V_MIN_TYP_MAX", "V_PROGRAM",       "V_CLASS_DECL",
+   "V_NULL",          "V_CLASS_NEW",   "V_CONSTRUCTOR",   "V_SUPER_CALL",
+   "V_IMPORT_DECL",   "V_NAMESPACE",   "V_DEFPARAM",
 };
 
 static const change_allowed_t change_allowed[] = {
@@ -321,8 +350,8 @@ object_class_t vlog_object = {
    .tag            = OBJECT_TAG_VLOG,
    .last_kind      = V_LAST_NODE_KIND,
    .has_loc        = true,
-   .gc_roots       = { V_MODULE, V_PRIMITIVE },
-   .gc_num_roots   = 2
+   .gc_roots       = { V_MODULE, V_PRIMITIVE, V_PROGRAM, V_PACKAGE },
+   .gc_num_roots   = 4,
 };
 
 struct _vlog_node {
@@ -671,6 +700,16 @@ void vlog_set_ival(vlog_node_t v, int64_t i)
    lookup_item(&vlog_object, v, I_IVAL)->ival = i;
 }
 
+vlog_flags_t vlog_flags(vlog_node_t t)
+{
+   return lookup_item(&vlog_object, t, I_FLAGS)->ival;
+}
+
+void vlog_set_flags(vlog_node_t v, vlog_flags_t mask)
+{
+   lookup_item(&vlog_object, v, I_FLAGS)->ival |= mask;
+}
+
 void vlog_visit(vlog_node_t v, vlog_visit_fn_t fn, void *context)
 {
    vlog_visit_only(v, fn, context, V_LAST_NODE_KIND);
