@@ -144,25 +144,27 @@ void vlog_symtab_pop(vlog_symtab_t *st)
 {
    assert(st->top != NULL);
 
-   if (st->top->parent == NULL) {
-      for (int i = 0; i < st->top->deferred.count; i++) {
-         vlog_node_t v = st->top->deferred.items[i];
+   unsigned wptr = 0;
+   for (int i = 0; i < st->top->deferred.count; i++) {
+      vlog_node_t v = st->top->deferred.items[i];
+      vlog_node_t d = vlog_symtab_query(st, vlog_ident(v));
+      if (d != NULL)
+         vlog_set_ref(v, d);
+      else
+         st->top->deferred.items[wptr++] = v;
+   }
+   ATRIM(st->top->deferred, wptr);
+
+   for (int i = 0; i < st->top->deferred.count; i++) {
+      vlog_node_t v = st->top->deferred.items[i];
+      if (st->top->parent == NULL || is_top_level(st->top->container)) {
          ident_t name = vlog_ident(v);
-         error_at(vlog_loc(v), "no visible declaration for '%s'", istr(name));
+         error_at(vlog_loc(v), "no visible declaration for '%pi'", name);
          vlog_symtab_poison(st, name);
       }
+      else
+         APUSH(st->top->parent->deferred, v);
    }
-   else {
-      for (int i = 0; i < st->top->deferred.count; i++) {
-         vlog_node_t v = st->top->deferred.items[i];
-         const vlog_symbol_t *sym = symbol_for(st, vlog_ident(v));
-         if (sym != NULL)
-            vlog_set_ref(v, sym->node);
-         else
-            APUSH(st->top->parent->deferred, v);
-      }
-   }
-   ACLEAR(st->top->deferred);
 
    st->top = st->top->parent;
 }
@@ -172,17 +174,19 @@ void vlog_symtab_set_implicit(vlog_symtab_t *st, vlog_net_kind_t kind)
    st->implicit = kind;
 }
 
-static void lookup_struct_member(vlog_symtab_t *st, vlog_node_t v)
+static void lookup_member(vlog_symtab_t *st, vlog_node_t v)
 {
    vlog_node_t prefix = vlog_value(v), type = vlog_get_type(prefix);
    if (type == NULL) {  // Was earlier error
       assert(error_count() > 0);
       return;
    }
-   else if (vlog_kind(type) != V_STRUCT_DECL) {
+
+   const vlog_kind_t kind = vlog_kind(type);
+   if (kind != V_STRUCT_DECL && kind != V_CLASS_DECL) {
       diag_t *d = diag_new(DIAG_ERROR, vlog_loc(prefix));
       if (vlog_kind(prefix) == V_REF)
-         diag_printf(d, "'%s'", istr(vlog_ident(prefix)));
+         diag_printf(d, "'%pi'", vlog_ident(prefix));
       else
          diag_printf(d, "prefix");
       diag_printf(d, " is not a struct or class");
@@ -202,7 +206,11 @@ static void lookup_struct_member(vlog_symtab_t *st, vlog_node_t v)
    }
 
    diag_t *d = diag_new(DIAG_ERROR, vlog_loc(v));
-   diag_printf(d, "struct has no field named '%s'", istr(id));
+   if (kind == V_STRUCT_DECL)
+      diag_printf(d, "struct ");
+   else
+      diag_printf(d, "class '%pi' ", vlog_ident(type));
+   diag_printf(d, "has no field named '%pi'", id);
    diag_hint(d, vlog_loc(type), "struct declared here");
    diag_emit(d);
 }
@@ -215,8 +223,8 @@ void vlog_symtab_lookup(vlog_symtab_t *st, vlog_node_t v)
    assert(!loc_invalid_p(vlog_loc(v)));
 
    switch (vlog_kind(v)) {
-   case V_STRUCT_REF:
-      lookup_struct_member(st, v);
+   case V_MEMBER_REF:
+      lookup_member(st, v);
       return;
    default:
       break;
@@ -238,7 +246,6 @@ void vlog_symtab_lookup(vlog_symtab_t *st, vlog_node_t v)
 
       vlog_set_ref(v, decl);
 
-      assert(vlog_kind(st->top->container) == V_MODULE);
       vlog_add_decl(st->top->container, decl);
 
       fresh_symbol_for(st, name)->node = decl;
@@ -252,7 +259,7 @@ void vlog_symtab_lookup(vlog_symtab_t *st, vlog_node_t v)
          APUSH(st->top->deferred, v);
          break;
       default:
-         error_at(vlog_loc(v), "no visible declaration for '%s'", istr(name));
+         error_at(vlog_loc(v), "no visible declaration for '%pi'", name);
          vlog_symtab_poison(st, name);
       }
    }
@@ -270,24 +277,25 @@ void vlog_symtab_put(vlog_symtab_t *st, vlog_node_t v)
    if (sym->node == POISON_NODE)
       return;
 
-   if (sym->node != NULL && vlog_kind(v) == V_PORT_DECL) {
+   if (sym->node != NULL) {
       switch (vlog_kind(sym->node)) {
       case V_VAR_DECL:
       case V_NET_DECL:
-         if (!vlog_has_ref(v)) {
-            vlog_set_ref(v, sym->node);
-            sym->node = v;
-            return;
+         switch (vlog_kind(v)) {
+         case V_PORT_DECL:
+         case V_TF_PORT_DECL:
+            if (!vlog_has_ref(v)) {
+               vlog_set_ref(v, sym->node);
+               sym->node = v;
+               return;
+            }
+            break;
+         default:
+            break;
          }
          break;
-      default:
-         break;
-      }
-   }
-
-   if (sym->node != NULL) {
-      switch (vlog_kind(sym->node)) {
       case V_PORT_DECL:
+      case V_TF_PORT_DECL:
          if (!vlog_has_ref(sym->node)) {
             switch (vlog_kind(v)) {
             case V_VAR_DECL:
@@ -303,9 +311,9 @@ void vlog_symtab_put(vlog_symtab_t *st, vlog_node_t v)
       }
 
       diag_t *d = diag_new(DIAG_ERROR, vlog_loc(v));
-      diag_printf(d, "duplicate declaration of %s", istr(name));
-      diag_hint(d, vlog_loc(sym->node), "%s was previously declared here",
-                istr(name));
+      diag_printf(d, "duplicate declaration of '%pi'", name);
+      diag_hint(d, vlog_loc(sym->node), "'%pi' was previously declared here",
+                name);
       diag_hint(d, vlog_loc(v), "duplicate declaration");
       diag_suppress(d, name == well_known(W_ERROR) || st->top->suppress);
       diag_emit(d);
