@@ -18,8 +18,10 @@
 #include "util.h"
 #include "diag.h"
 #include "jit/jit.h"
+#include "option.h"
 #include "rt/model.h"
-#include "rt/random.h"
+#include "svrand.h"
+#include "thread.h"
 #include "vpi/vpi-priv.h"
 
 #include <assert.h>
@@ -64,13 +66,15 @@ static void format_number(vpiHandle it, char radix, int fwidth)
          s_vpi_value argval = { .format = vpiDecStrVal };
          vpi_get_value(arg, &argval);
 
-         const int nbits = vpi_get(vpiSize, arg);
-         const int dmax = calc_dec_size(nbits, false);
+         if (!vpi_chk_error(NULL)) {
+            const int nbits = vpi_get(vpiSize, arg);
+            const int dmax = calc_dec_size(nbits, false);
 
-         if (dmax > strlen(argval.value.str))
-            printf("%*s", dmax, argval.value.str);
-         else
-            fputs(argval.value.str, stdout);
+            if (dmax > strlen(argval.value.str))
+               printf("%*s", dmax, argval.value.str);
+            else
+               fputs(argval.value.str, stdout);
+         }
       }
       break;
    case 'x':
@@ -79,7 +83,8 @@ static void format_number(vpiHandle it, char radix, int fwidth)
          s_vpi_value argval = { .format = vpiHexStrVal };
          vpi_get_value(arg, &argval);
 
-         fputs(argval.value.str, stdout);
+         if (!vpi_chk_error(NULL))
+            fputs(argval.value.str, stdout);
       }
       break;
    case 'b':
@@ -87,7 +92,17 @@ static void format_number(vpiHandle it, char radix, int fwidth)
          s_vpi_value argval = { .format = vpiBinStrVal };
          vpi_get_value(arg, &argval);
 
-         fputs(argval.value.str, stdout);
+         if (!vpi_chk_error(NULL))
+            fputs(argval.value.str, stdout);
+      }
+      break;
+   case 'f':
+      {
+         s_vpi_value argval = { .format = vpiRealVal };
+         vpi_get_value(arg, &argval);
+
+         if (!vpi_chk_error(NULL))
+            printf("%f", argval.value.real);
       }
       break;
    }
@@ -104,6 +119,8 @@ static void format_char(vpiHandle it, int fwidth)
    s_vpi_value argval = { .format = vpiDecStrVal };
    vpi_get_value(arg, &argval);
 
+   s_vpi_error_info ei;
+   vpi_chk_error(&ei);
    const char ch = atoi(argval.value.str);
    fputc(ch, stdout);
 
@@ -134,6 +151,7 @@ static void interpret_format(const char *fmt, vpiHandle it)
          case 'x':
          case 'h':
          case 't':
+         case 'f':
             format_number(it, *p, fwidth);
             break;
          case 'c':
@@ -181,10 +199,30 @@ static void verilog_printf(void)
       arg = vpi_scan(it);
    }
 
-   // TODO: iterate remaining arguments
+   while (arg != NULL) {
+      const bool is_null = vpi_get(vpiType, arg) == vpiOperation
+         && vpi_get(vpiOpType, arg) == vpiNullOp;
+
+      if (is_null)
+         fputc(' ', stdout);
+      else {
+         s_vpi_value argval = { .format = vpiDecStrVal };
+         vpi_get_value(arg, &argval);
+
+         const int nbits = vpi_get(vpiSize, arg);
+         const int dmax = calc_dec_size(nbits, false);
+
+         if (dmax > strlen(argval.value.str))
+            printf("%*s", dmax, argval.value.str);
+         else
+            fputs(argval.value.str, stdout);
+      }
+
+      vpi_release_handle(arg);
+      arg = vpi_scan(it);
+   }
 
  release_handles:
-   vpi_release_handle(it);
    vpi_release_handle(call);
 }
 
@@ -222,6 +260,49 @@ static PLI_INT32 monitor_tf(PLI_BYTE8 *userdata)
    return 0;
 }
 
+static PLI_INT32 readmemh_tf(PLI_BYTE8 *userdata)
+{
+   vpiHandle callh = vpi_handle(vpiSysTfCall, 0);
+   vpiHandle argv = vpi_iterate(vpiArgument, callh);
+
+   vpiHandle file_arg = vpi_scan(argv);
+   vpiHandle mem_arg = vpi_scan(argv);
+
+   s_vpi_value file = { .format = vpiStringVal };
+   vpi_get_value(file_arg, &file);
+   vpi_release_handle(file_arg);
+
+   FILE *f = fopen(file.value.str, "r");
+   if (f == NULL)
+      jit_msg(NULL, DIAG_FATAL, "failed to open %s: %s",
+              file.value.str, last_os_error());
+
+   char *line LOCAL = NULL;
+   size_t line_len = 0, index = 0;
+   while (getline(&line, &line_len, f) != -1) {
+      char *savep = NULL;
+      char *tok = strtok_r(line, " \t\r\n", &savep);
+      if (tok == NULL)
+         continue;
+
+      vpiHandle elem = vpi_handle_by_index(mem_arg, index++);
+      if (elem == NULL)
+         break;
+
+      s_vpi_value val = { .format = vpiHexStrVal, .value.str = tok };
+      vpi_put_value(elem, &val, NULL, vpiNoDelay);
+
+      vpi_release_handle(elem);
+   }
+
+   fclose(f);
+
+   vpi_release_handle(mem_arg);
+   vpi_release_handle(argv);
+   vpi_release_handle(callh);
+   return 0;
+}
+
 static PLI_INT32 time_tf(PLI_BYTE8 *userdata)
 {
    rt_model_t *m = get_model();
@@ -249,16 +330,44 @@ static PLI_INT32 time_tf(PLI_BYTE8 *userdata)
 
 static PLI_INT32 random_tf(PLI_BYTE8 *userdata)
 {
-   s_vpi_value result = {
-      .format = vpiIntVal,
-      .value = { .integer = get_random() },
-   };
+   static __thread int32_t i_seed;
+   INIT_ONCE(i_seed = opt_get_int(OPT_RANDOM_SEED));
+
+   int32_t a_seed = i_seed;
 
    vpiHandle call = vpi_handle(vpiSysTfCall, NULL);
    assert(call != NULL);
 
+   vpiHandle argv = vpi_iterate(vpiArgument, call);
+
+   vpiHandle seed = vpi_scan(argv);
+   if (seed != NULL) {
+      s_vpi_value val = { .format = vpiIntVal };
+      vpi_get_value(seed, &val);
+
+      a_seed = val.value.integer;
+   }
+
+   vpi_release_handle(argv);
+
+   s_vpi_value result = {
+      .format = vpiIntVal,
+      .value = { .integer = rtl_dist_uniform(&a_seed, INT32_MIN, INT32_MAX) },
+   };
+
    vpi_put_value(call, &result, NULL, 0);
 
+   if (seed != NULL) {
+      s_vpi_value next_seed = {
+         .format = vpiIntVal,
+         .value = { .integer = a_seed },
+      };
+      vpi_put_value(seed, &next_seed, NULL, 0);
+      vpi_release_handle(seed);
+   }
+   else
+      i_seed = a_seed;
+
    vpi_release_handle(call);
    return 0;
 }
@@ -289,6 +398,11 @@ static s_vpi_systf_data builtins[] = {
       .tfname = "$monitor",
       .calltf = monitor_tf
    },
+   {
+      .type   = vpiSysTask,
+      .tfname = "$readmemh",
+      .calltf = readmemh_tf
+   },
    {
       .type        = vpiSysFunc,
       .tfname      = "$time",
