@@ -21,6 +21,8 @@
 #include "diag.h"
 #include "jit/jit.h"
 #include "lib.h"
+#include "lower.h"
+#include "mir/mir-unit.h"
 #include "option.h"
 #include "phase.h"
 #include "rt/model.h"
@@ -156,7 +158,7 @@ START_TEST(test_copy1)
 
    tree_t func1 = get_decl(sub1_i, "DOUBLE");
    tree_t func2 = get_decl(sub2_i, "DOUBLE");
-   fail_if(func1 == func2);   // Should copy functions
+   fail_unless(func1 == func2);   // Should not copy functions
 
    tree_t var1 = get_decl(sub1_i, "GLOBAL");
    tree_t var2 = get_decl(sub2_i, "GLOBAL");
@@ -817,8 +819,8 @@ START_TEST(test_tc3138)
    tree_t b0 = tree_stmt(e, 0);
    fail_unless(tree_ident(b0) == ident_new("C05S02B02X00P02N01I03138ENT"));
    fail_unless(tree_stmts(b0) == 2);
-   fail_unless(tree_kind(tree_stmt(b0, 0)) == T_BLOCK);
-   fail_unless(tree_kind(tree_stmt(b0, 1)) == T_PROCESS);
+   fail_unless(tree_kind(tree_stmt(b0, 0)) == T_PROCESS);
+   fail_unless(tree_kind(tree_stmt(b0, 1)) == T_BLOCK);
 
    fail_if_errors();
 }
@@ -857,7 +859,7 @@ START_TEST(test_tc846)
    fail_unless(tree_ident(b0) == ident_new("C01S03B01X00P08N01I00846ENT"));
    fail_unless(tree_stmts(b0) == 2);
 
-   tree_t a1 = tree_stmt(b0, 0);
+   tree_t a1 = tree_stmt(b0, 1);
    fail_unless(tree_ident(a1) == ident_new("A1"));
    fail_unless(tree_stmts(a1) == 1);
 
@@ -1351,8 +1353,9 @@ START_TEST(test_generic1)
    fail_unless(tree_kind(p1s0) == T_SIGNAL_ASSIGN);
 
    tree_t one = tree_value(tree_waveform(p1s0, 0));
-   fail_unless(tree_kind(one) == T_LITERAL);
-   fail_unless(tree_ival(one) == 1);
+   fail_unless(tree_kind(one) == T_REF);
+   // XXX: used to be folded
+   //   fail_unless(tree_ival(one) == 1);
 
    tree_t p2s0 = tree_stmt(tree_stmt(u, 1), 0);
    fail_unless(tree_kind(p2s0) == T_ASSERT);
@@ -1385,16 +1388,18 @@ START_TEST(test_generic2)
    fail_unless(tree_kind(p1s0) == T_SIGNAL_ASSIGN);
 
    tree_t two = tree_value(tree_waveform(p1s0, 0));
-   fail_unless(tree_kind(two) == T_LITERAL);
-   fail_unless(tree_ival(two) == 2);
+   fail_unless(tree_kind(two) == T_REF);
+   // XXX: used to be folded
+   //   fail_unless(tree_ival(two) == 2);
 
    tree_t p2s0 = tree_stmt(tree_stmt(u, 1), 0);
    fail_unless(tree_kind(p2s0) == T_ASSERT);
 
    tree_t value = tree_value(p2s0);
    fail_unless(tree_kind(value) == T_REF);
-   fail_unless(tree_ident(value) == ident_new("FALSE"));
-   fail_unless(tree_kind(tree_ref(value)) == T_ENUM_LIT);
+   // XXX: used to be folded
+   //   fail_unless(tree_ident(value) == ident_new("FALSE"));
+   //   fail_unless(tree_kind(tree_ref(value)) == T_ENUM_LIT);
 
    fail_if_errors();
 }
@@ -1404,11 +1409,6 @@ START_TEST(test_issue669)
 {
    input_from_file(TESTDIR "/elab/issue669.vhd");
 
-   const error_t expect[] = {
-      { -1, NULL }
-   };
-   expect_errors(expect);
-
    tree_t e = run_elab();
    fail_if(e == NULL);
 
@@ -1631,7 +1631,7 @@ START_TEST(test_issue759)
    mir_context_t *mc = get_mir();
    unit_registry_t *ur = get_registry();
    cover_data_t *cover = cover_data_init(COVER_MASK_ALL, 0, 0);
-   jit_t *jit = jit_new(ur, mc, cover);
+   jit_t *jit = jit_new(ur, mc);
 
    rt_model_t *m = model_new(jit, cover);
    tree_t e = elab(tree_to_object(a), jit, ur, mc, cover, NULL, m);
@@ -1747,11 +1747,12 @@ START_TEST(test_vlog1)
    const error_t expect[] = {
       { 12, "name of Verilog module sub1 in library unit WORK.SUB1 does "
         "not match name SUB1 in module instance u2" },
-      { 13, "expected 2 port connections for module sub1 but found 1" },
+      { 13, "missing port connection for 'y'" },
       { 14, "module bad not found in library WORK" },
       { 15, "missing value for parameter p" },
       { 16, "module sub1 has 1 parameter but 2 values given" },
       { 17, "missing port connection for 'x'" },
+      { 17, "port 'foo' not found in sub1" },
       { 18, "'z' cannot be driven by continuous assignment from port 'y'" },
       { 19, "expression cannot be driven by continuous assignment from "
         "port 'y'" },
@@ -1766,7 +1767,7 @@ START_TEST(test_vlog1)
 
    mir_context_t *mc = get_mir();
    unit_registry_t *ur = get_registry();
-   jit_t *j = jit_new(ur, mc, NULL);
+   jit_t *j = jit_new(ur, mc);
    rt_model_t *m = model_new(j, NULL);
 
    tree_t top = elab(obj, j, ur, mc, NULL, NULL, m);
@@ -1955,7 +1956,7 @@ START_TEST(test_issue1012)
 
    mir_context_t *mc = get_mir();
    unit_registry_t *ur = get_registry();
-   jit_t *jit = jit_new(ur, mc, NULL);
+   jit_t *jit = jit_new(ur, mc);
    cover_data_t *cover = cover_data_init(COVER_MASK_TOGGLE, 0, 0);
    rt_model_t *m = model_new(jit, NULL);
 
@@ -2115,6 +2116,113 @@ START_TEST(test_mixed2)
 }
 END_TEST
 
+START_TEST(test_clone2)
+{
+   set_standard(STD_19);
+
+   input_from_file(TESTDIR "/elab/clone2.vhd");
+
+   tree_t e = run_elab();
+   fail_if(e == NULL);
+
+   mir_context_t *mc = get_mir();
+   unit_registry_t *ur = get_registry();
+
+   tree_t top = tree_stmt(e, 0);
+
+   tree_t u1 = tree_stmt(top, 0);
+   fail_unless(tree_kind(u1) == T_BLOCK);
+   fail_unless(tree_ident(u1) == ident_new("U1"));
+
+   tree_t u2 = tree_stmt(top, 1);
+   fail_unless(tree_kind(u2) == T_BLOCK);
+   fail_unless(tree_ident(u2) == ident_new("U2"));
+
+   fail_unless(tree_port(u1, 0) == tree_port(u2, 0));
+
+   tree_t g1_1 = tree_stmt(top, 2);
+   fail_unless(tree_kind(g1_1) == T_BLOCK);
+   fail_unless(tree_ident(g1_1) == ident_new("G1(1)"));
+
+   tree_t g1_2 = tree_stmt(top, 3);
+   fail_unless(tree_kind(g1_2) == T_BLOCK);
+   fail_unless(tree_ident(g1_2) == ident_new("G1(2)"));
+
+   fail_unless(tree_stmt(g1_1, 0) == tree_stmt(g1_2, 0));
+
+   ident_t g1psym1 = ident_new("WORK.CLONE2.G1(1).P");
+   ident_t g1psym2 = ident_new("WORK.CLONE2.G1(2).P");
+   (void)unit_registry_get(ur, g1psym1);
+   (void)unit_registry_get(ur, g1psym2);
+   ck_assert_ptr_nonnull(mir_get_unit(mc, g1psym1));
+   ck_assert_ptr_null(mir_get_unit(mc, g1psym2));
+
+   tree_t u3 = tree_stmt(top, 5);
+   fail_unless(tree_kind(u3) == T_BLOCK);
+   fail_unless(tree_ident(u3) == ident_new("U3"));
+
+   tree_t u4 = tree_stmt(top, 6);
+   fail_unless(tree_kind(u4) == T_BLOCK);
+   fail_unless(tree_ident(u4) == ident_new("U4"));
+
+   fail_unless(tree_port(u3, 0) == tree_port(u4, 0));
+
+   tree_t u5 = tree_stmt(top, 7);
+   fail_unless(tree_kind(u5) == T_BLOCK);
+   fail_unless(tree_ident(u5) == ident_new("U5"));
+
+   tree_t u6 = tree_stmt(top, 8);
+   fail_unless(tree_kind(u6) == T_BLOCK);
+   fail_unless(tree_ident(u6) == ident_new("U6"));
+
+   fail_if(tree_port(u5, 0) == tree_port(u6, 0));
+
+   tree_t u7 = tree_stmt(top, 9);
+   fail_unless(tree_kind(u7) == T_BLOCK);
+   fail_unless(tree_ident(u7) == ident_new("U7"));
+
+   tree_t u7h = tree_decl(u7, 0);
+   fail_unless(tree_kind(u7h) == T_HIER);
+   fail_unless(tree_ident(u7h) == ident_new("WORK.CLONE2.U7"));
+   ck_assert_ident_eq(tree_ident2(u7h), "WORK.CLONE2.U1.SUB1_ENT");
+
+   ident_t u7psym1 = ident_new("WORK.CLONE2.U1.SUB1_ENT.PROC");
+   ident_t u7psym2 = ident_new("WORK.CLONE2.U7.PROC");
+   (void)unit_registry_get(ur, u7psym1);
+   (void)unit_registry_get(ur, u7psym2);
+   ck_assert_ptr_nonnull(mir_get_unit(mc, u7psym1));
+   ck_assert_ptr_null(mir_get_unit(mc, u7psym2));
+
+   fail_if(tree_port(u5, 0) == tree_port(u1, 0));
+
+   fail_if_errors();
+}
+END_TEST
+
+START_TEST(test_issue1333)
+{
+   input_from_file(TESTDIR "/elab/issue1333.v");
+
+   const error_t expect[] = {
+      { 11, "specify blocks are not currently supported and will be ignored" },
+      { 20, "missing port connection for 'Y'" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   tree_t e = run_elab();
+   fail_if(e == NULL);   // Warning only
+
+   tree_t b0 = tree_stmt(e, 0);
+   ck_assert_int_eq(tree_decls(b0), 2);
+   ck_assert_int_eq(tree_stmts(b0), 2);
+   ck_assert(tree_kind(tree_stmt(b0, 0)) == T_BLOCK);
+   ck_assert(tree_kind(tree_stmt(b0, 1)) == T_VERILOG);
+
+   check_expected_errors();
+}
+END_TEST
+
 Suite *get_elab_tests(void)
 {
    Suite *s = suite_create("elab");
@@ -2226,6 +2334,8 @@ Suite *get_elab_tests(void)
    tcase_add_test(tc, test_issue1204);
    tcase_add_test(tc, test_clone1);
    tcase_add_test(tc, test_mixed2);
+   tcase_add_test(tc, test_clone2);
+   tcase_add_test(tc, test_issue1333);
    suite_add_tcase(s, tc);
 
    return s;
