@@ -83,7 +83,7 @@ START_TEST(test_cfold)
 
    unit_registry_t *ur = get_registry();
    mir_context_t *mc = get_mir();
-   jit_t *jit = jit_new(ur, mc, NULL);
+   jit_t *jit = jit_new(ur, mc);
    simplify_local(a, jit, ur, mc);
    jit_free(jit);
 
@@ -197,7 +197,7 @@ START_TEST(test_proc)
 
    mir_context_t *mc = get_mir();
    unit_registry_t *ur = get_registry();
-   jit_t *jit = jit_new(ur, mc, NULL);
+   jit_t *jit = jit_new(ur, mc);
    simplify_local(a, jit, ur, mc);
    jit_free(jit);
 
@@ -280,7 +280,7 @@ START_TEST(test_args)
 
    mir_context_t *mc = get_mir();
    unit_registry_t *ur = get_registry();
-   jit_t *jit = jit_new(ur, mc, NULL);
+   jit_t *jit = jit_new(ur, mc);
    simplify_local(a, jit, ur, mc);
    jit_free(jit);
 
@@ -332,7 +332,7 @@ START_TEST(test_ffold)
 
    mir_context_t *mc = get_mir();
    unit_registry_t *ur = get_registry();
-   jit_t *jit = jit_new(ur, mc, NULL);
+   jit_t *jit = jit_new(ur, mc);
    simplify_global(b, NULL, jit, ur, mc);
    jit_free(jit);
    fail_if_errors();
@@ -382,7 +382,7 @@ START_TEST(test_ffold2)
 
    mir_context_t *mc = get_mir();
    unit_registry_t *ur = get_registry();
-   jit_t *jit = jit_new(ur, mc, NULL);
+   jit_t *jit = jit_new(ur, mc);
    simplify_global(b, NULL, jit, ur, mc);
    jit_free(jit);
    fail_if_errors();
@@ -400,7 +400,7 @@ START_TEST(test_issue49)
 
    mir_context_t *mc = get_mir();
    unit_registry_t *ur = get_registry();
-   jit_t *jit = jit_new(ur, mc, NULL);
+   jit_t *jit = jit_new(ur, mc);
    simplify_local(a, jit, ur, mc);
    jit_free(jit);
 }
@@ -415,7 +415,7 @@ START_TEST(test_issue155)
 
    mir_context_t *mc = get_mir();
    unit_registry_t *ur = get_registry();
-   jit_t *jit = jit_new(ur, mc, NULL);
+   jit_t *jit = jit_new(ur, mc);
    simplify_global(p, NULL, jit, ur, mc);
    jit_free(jit);
 
@@ -457,7 +457,7 @@ START_TEST(test_context)
 
    mir_context_t *mc = get_mir();
    unit_registry_t *ur = get_registry();
-   jit_t *jit = jit_new(ur, mc, NULL);
+   jit_t *jit = jit_new(ur, mc);
    simplify_local(e, jit, ur, mc);
    jit_free(jit);
 
@@ -1071,7 +1071,7 @@ START_TEST(test_order1)
    tree_t p = parse_check_and_simplify(T_PACKAGE, T_PACKAGE);
 
    tree_t x = get_decl(p, "X");
-   fail_unless(tree_kind(tree_value(x)) == T_FCALL);
+   fail_unless(folded_b(tree_value(x), false));
 
    check_expected_errors();
 }
@@ -1631,7 +1631,7 @@ START_TEST(test_casearray1)
    fail_unless(tree_kind(s0) == T_CASE);
 
    for (int i = 0; i < 4; i++) {
-      tree_t n0 = tree_name(tree_assoc(tree_stmt(s0, i), 0));
+      tree_t n0 = tree_name(tree_choice(tree_stmt(s0, i), 0));
       fail_unless(tree_kind(n0) == T_STRING);
       fail_unless(tree_chars(n0) == 8);
    }
@@ -1886,15 +1886,10 @@ START_TEST(test_issue1182)
    tree_t a = parse_check_and_simplify(T_PACKAGE, T_PACKAGE, T_ENTITY, T_ARCH);
 
    tree_t p_proc = tree_stmt(a, 0);
-   fail_unless(tree_stmts(p_proc) == 4);
+   fail_unless(tree_stmts(p_proc) == 2);
 
    tree_t a0 = tree_stmt(p_proc, 0);
-   fail_unless(tree_kind(a0) == T_ASSERT);
-   fail_unless(tree_kind(tree_value(a0)) == T_FCALL);  // Could be folded
-
-   tree_t a2 = tree_stmt(p_proc, 2);
-   fail_unless(tree_kind(a2) == T_ASSERT);
-   fail_unless(tree_kind(tree_value(a2)) == T_FCALL);  // Could be folded
+   fail_unless(tree_kind(a0) == T_SIGNAL_ASSIGN);
 
    fail_if_errors();
 }
@@ -1951,10 +1946,31 @@ START_TEST(test_issue1347)
    tree_t a = parse_check_and_simplify(T_ENTITY, T_ARCH);
 
    tree_t c = tree_stmt(tree_stmt(a, 0), 0);
-   fail_unless(folded_i(tree_name(tree_assoc(tree_stmt(c, 0), 0)), 0));
-   fail_unless(folded_i(tree_name(tree_assoc(tree_stmt(c, 1), 0)), 42));
-   fail_unless(folded_i(tree_name(tree_assoc(tree_stmt(c, 2), 0)), 3));
-   fail_unless(folded_i(tree_name(tree_assoc(tree_stmt(c, 2), 1)), -1024));
+   fail_unless(folded_i(tree_name(tree_choice(tree_stmt(c, 0), 0)), 0));
+   fail_unless(folded_i(tree_name(tree_choice(tree_stmt(c, 1), 0)), 42));
+   fail_unless(folded_i(tree_name(tree_choice(tree_stmt(c, 2), 0)), 3));
+   fail_unless(folded_i(tree_name(tree_choice(tree_stmt(c, 2), 1)), -1024));
+
+   fail_if_errors();
+}
+END_TEST
+
+START_TEST(test_issue1353)
+{
+   set_standard(STD_08);
+   input_from_file(TESTDIR "/simp/issue1353.vhd");
+
+   tree_t a = parse_check_and_simplify(T_ENTITY, T_ARCH);
+
+   tree_t c = tree_stmt(tree_stmt(a, 0), 0);
+   fail_unless(folded_i(tree_name(tree_choice(tree_stmt(c, 0), 0)), 0));
+   fail_unless(folded_i(tree_name(tree_choice(tree_stmt(c, 0), 1)), 1));
+   tree_t r1 = tree_range(tree_choice(tree_stmt(c, 1), 0), 0);
+   fail_unless(folded_i(tree_left(r1), 2));
+   fail_unless(folded_i(tree_right(r1), 3));
+   tree_t r2 = tree_range(tree_choice(tree_stmt(c, 2), 0), 0);
+   fail_unless(folded_i(tree_left(r2), 4));
+   fail_unless(folded_i(tree_right(r2), 5));
 
    fail_if_errors();
 }
@@ -2036,6 +2052,7 @@ Suite *get_simp_tests(void)
    tcase_add_test(tc_core, test_issue1239);
    tcase_add_test(tc_core, test_issue1318);
    tcase_add_test(tc_core, test_issue1347);
+   tcase_add_test(tc_core, test_issue1353);
    suite_add_tcase(s, tc_core);
 
    return s;
