@@ -15,20 +15,40 @@
 //  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 //
 
+#include "test_util.h"
 #include "phase.h"
 #include "common.h"
 #include "ident.h"
 #include "lib.h"
 #include "object.h"
+#include "option.h"
 #include "prim.h"
 #include "scan.h"
-#include "test_util.h"
+#include "test_mir.h"
 #include "vlog/vlog-node.h"
 #include "vlog/vlog-number.h"
 #include "vlog/vlog-phase.h"
 
 #include <inttypes.h>
 
+static vlog_node_t do_parse_only(vlog_kind_t kind)
+{
+   vlog_node_t v = vlog_parse();
+   ck_assert_ptr_nonnull(v);
+   ck_assert_msg(vlog_kind(v) == kind, "expected %s but have %s",
+                 vlog_kind_str(kind), vlog_kind_str(vlog_kind(v)));
+   return v;
+}
+
+static vlog_node_t do_parse_check(vlog_kind_t kind)
+{
+   const int base_errors = error_count();
+   vlog_node_t v = do_parse_only(kind);
+   ck_assert_int_eq(error_count(), base_errors);
+   vlog_check(v);
+   return v;
+}
+
 START_TEST(test_dff)
 {
    input_from_file(TESTDIR "/vlog/dff.v");
@@ -38,7 +58,7 @@ START_TEST(test_dff)
    fail_unless(vlog_kind(m) == V_MODULE);
    fail_unless(vlog_stmts(m) == 1);
    fail_unless(vlog_ports(m) == 4);
-   fail_unless(vlog_decls(m) == 5);
+   fail_unless(vlog_decls(m) == 4);
 
    vlog_node_t p0 = vlog_port(m, 0);
    fail_unless(vlog_kind(p0) == V_REF);
@@ -70,10 +90,6 @@ START_TEST(test_dff)
    fail_unless(vlog_ident(d3) == ident_new("q"));
    fail_unless(vlog_ident2(d3) == ident_new("Q"));
 
-   vlog_node_t d4 = vlog_decl(m, 4);
-   fail_unless(vlog_kind(d4) == V_VAR_DECL);
-   fail_unless(vlog_ident(d4) == ident_new("q"));
-
    vlog_node_t a = vlog_stmt(m, 0);
    fail_unless(vlog_kind(a) == V_ALWAYS);
    fail_unless(vlog_stmts(a) == 1);
@@ -110,7 +126,7 @@ END_TEST
 START_TEST(test_simple_sem)
 {
    const error_t expect[] = {
-      {  7, "duplicate declaration of x" },
+      {  7, "duplicate declaration of 'x'" },
       { 13, "no visible declaration for 'qq'" },
       { 13, "no visible declaration for 'dd'" },
       { 19, "'r' cannot be driven by continuous assignment" },
@@ -121,12 +137,38 @@ START_TEST(test_simple_sem)
 
    input_from_file(TESTDIR "/vlog/simple_sem.v");
 
-   for (int i = 0; i < 4; i++) {
+   for (int i = 0; i < 3; i++)
+      do_parse_only(V_MODULE);
+
+   do_parse_check(V_MODULE);
+
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
+START_TEST(test_ports1)
+{
+   input_from_file(TESTDIR "/vlog/ports1.v");
+
+   const error_t expect[] = {
+      {  4, "duplicate declaration of 'y'" },
+      {  0, "duplicate declaration" },
+      {  0, "'y' was previously declared here" },
+      { 11, "no visible declaration for 'y'" },
+      { 19, "duplicate declaration of 'x'" },
+      { 22, "duplicate declaration of 'y'" },
+      { 41, "'o' does not appear in module port list" },
+      { 45, "'i1' does not appear in module port list" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   for (int i = 0; i < 7; i++) {
       vlog_node_t m = vlog_parse();
       fail_if(m == NULL);
       fail_unless(vlog_kind(m) == V_MODULE);
-
-      vlog_check(m);
    }
 
    fail_unless(vlog_parse() == NULL);
@@ -135,32 +177,20 @@ START_TEST(test_simple_sem)
 }
 END_TEST
 
-START_TEST(test_ports)
+START_TEST(test_ports2)
 {
+   input_from_file(TESTDIR "/vlog/ports2.v");
+
    const error_t expect[] = {
-      {  4, "duplicate declaration of y" },
-      { 11, "no visible declaration for 'y'" },
-      { 19, "duplicate declaration of x" },
-      { 22, "duplicate declaration of y" },
-      { 31, "'o3' cannot be assigned in a procedural block" },
-      { 43, "inconsistent dimensions for 'y'" },
-      { 44, "cannot reference net 'x' in constant expression" },
-      { 47, "missing port declaration for 'zz'" },
-      { 55, "'o' does not appear in module port list" },
-      { 59, "'i1' does not appear in module port list" },
+      {  5, "'o3' cannot be assigned in a procedural block" },
+      { 17, "inconsistent dimensions for 'y'" },
+      { 21, "missing port declaration for 'zz'" },
       { -1, NULL }
    };
    expect_errors(expect);
 
-   input_from_file(TESTDIR "/vlog/ports.v");
-
-   for (int i = 0; i < 10; i++) {
-      vlog_node_t m = vlog_parse();
-      fail_if(m == NULL);
-      fail_unless(vlog_kind(m) == V_MODULE);
-
-      vlog_check(m);
-   }
+   for (int i = 0; i < 5; i++)
+      do_parse_check(V_MODULE);
 
    fail_unless(vlog_parse() == NULL);
 
@@ -177,9 +207,9 @@ START_TEST(test_parse1)
    vlog_node_t m = vlog_parse();
    fail_if(m == NULL);
    fail_unless(vlog_kind(m) == V_MODULE);
-   ck_assert_int_eq(vlog_stmts(m), 39);
+   ck_assert_int_eq(vlog_stmts(m), 43);
    ck_assert_int_eq(vlog_ports(m), 0);
-   ck_assert_int_eq(vlog_decls(m), 36);
+   ck_assert_int_eq(vlog_decls(m), 38);
 
    vlog_node_t x = vlog_decl(m, 0);
    fail_unless(vlog_kind(x) == V_NET_DECL);
@@ -188,7 +218,7 @@ START_TEST(test_parse1)
 
    vlog_node_t xdt = vlog_type(x);
    fail_unless(vlog_kind(xdt) == V_DATA_TYPE);
-   fail_unless(vlog_subkind(xdt) == DT_LOGIC);
+   fail_unless(vlog_subkind(xdt) == DT_IMPLICIT);
    fail_unless(vlog_ranges(xdt) == 1);
 
    vlog_node_t xd0 = vlog_range(xdt, 0);
@@ -500,7 +530,7 @@ START_TEST(test_gate1)
    input_from_file(TESTDIR "/vlog/gate1.v");
 
    const error_t expect[] = {
-      {  6, "duplicate declaration of p1" },
+      {  6, "duplicate declaration of 'p1'" },
       { -1, NULL }
    };
    expect_errors(expect);
@@ -610,20 +640,14 @@ START_TEST(test_struct1)
    input_from_file(TESTDIR "/vlog/struct1.sv");
 
    const error_t expect[] = {
-      { 13, "duplicate declaration of a" },
-      { 18, "struct has no field named 'c'" },
-      { 19, "prefix is not a struct or class" },
-      { 20, "'t_pair' cannot be assigned in a procedural block" },
+      { 14, "duplicate declaration of 'a'" },
+      { 19, "struct has no field named 'c'" },
+      { 20, "prefix is not a struct or class" },
       { -1, NULL }
    };
    expect_errors(expect);
 
-   vlog_node_t m = vlog_parse();
-   fail_if(m == NULL);
-   fail_unless(vlog_kind(m) == V_MODULE);
-
-   vlog_check(m);
-
+   vlog_node_t m = do_parse_only(V_MODULE);
    ck_assert_int_eq(vlog_decls(m), 3);
 
    fail_unless(vlog_parse() == NULL);
@@ -632,24 +656,55 @@ START_TEST(test_struct1)
 }
 END_TEST
 
+START_TEST(test_struct2)
+{
+   input_from_file(TESTDIR "/vlog/struct2.sv");
+
+   const error_t expect[] = {
+      { 14, "'t_pair' cannot be assigned in a procedural block" },
+      { 15, "value of type 'integral' is not compatible with 'struct'" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   do_parse_check(V_MODULE);
+
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
 START_TEST(test_enum1)
 {
    input_from_file(TESTDIR "/vlog/enum1.sv");
 
    const error_t expect[] = {
-      {  5, "'a' is not a data type" },
+      {  6, "'a' is not a data type" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   vlog_node_t m = do_parse_only(V_MODULE);
+   ck_assert_int_eq(vlog_decls(m), 8);
+
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
+START_TEST(test_enum2)
+{
+   input_from_file(TESTDIR "/vlog/enum2.sv");
+
+   const error_t expect[] = {
       {  8, "cannot reference variable 'zz' in constant expression" },
       { -1, NULL }
    };
    expect_errors(expect);
 
-   vlog_node_t m = vlog_parse();
-   fail_if(m == NULL);
-   fail_unless(vlog_kind(m) == V_MODULE);
-
-   vlog_check(m);
-
-   ck_assert_int_eq(vlog_decls(m), 8);
+   do_parse_check(V_MODULE);
 
    fail_unless(vlog_parse() == NULL);
 
@@ -661,12 +716,7 @@ START_TEST(test_union1)
 {
    input_from_file(TESTDIR "/vlog/union1.sv");
 
-   vlog_node_t m = vlog_parse();
-   fail_if(m == NULL);
-   fail_unless(vlog_kind(m) == V_MODULE);
-
-   vlog_check(m);
-
+   vlog_node_t m = do_parse_check(V_MODULE);
    ck_assert_int_eq(vlog_decls(m), 2);
 
    fail_unless(vlog_parse() == NULL);
@@ -682,7 +732,7 @@ START_TEST(test_param1)
    set_default_keywords(VLOG_1800_2023);
 
    const error_t expect[] = {
-      {  4, "duplicate declaration of p1" },
+      {  4, "duplicate declaration of 'p1'" },
       { -1, NULL }
    };
    expect_errors(expect);
@@ -822,8 +872,6 @@ START_TEST(test_pp4)
    fail_if(m == NULL);
    fail_unless(vlog_kind(m) == V_MODULE);
 
-   vlog_check(m);
-
    fail_unless(vlog_parse() == NULL);
 
    check_expected_errors();
@@ -842,12 +890,7 @@ START_TEST(test_casename)
    expect_errors(expect);
 
    for (int i = 0; i < 2; i++) {
-      vlog_node_t m = vlog_parse();
-      fail_if(m == NULL);
-      fail_unless(vlog_kind(m) == V_MODULE);
-
-      vlog_check(m);
-
+      vlog_node_t m = do_parse_check(V_MODULE);
       lib_put_vlog(lib_work(), m);
    }
 
@@ -869,11 +912,7 @@ START_TEST(test_const1)
    };
    expect_errors(expect);
 
-   vlog_node_t m = vlog_parse();
-   fail_if(m == NULL);
-   fail_unless(vlog_kind(m) == V_MODULE);
-
-   vlog_check(m);
+   do_parse_check(V_MODULE);
 
    fail_unless(vlog_parse() == NULL);
 
@@ -886,17 +925,30 @@ START_TEST(test_case1)
    input_from_file(TESTDIR "/vlog/case1.v");
 
    const error_t expect[] = {
-      { 10, "no visible declaration for 'aaa'" },
-      { 15, "multiple default statements within a single case statement" },
+      { 11, "no visible declaration for 'aaa'" },
       { -1, NULL }
    };
    expect_errors(expect);
 
-   vlog_node_t m = vlog_parse();
-   fail_if(m == NULL);
-   fail_unless(vlog_kind(m) == V_MODULE);
+   do_parse_only(V_MODULE);
 
-   vlog_check(m);
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
+START_TEST(test_case2)
+{
+   input_from_file(TESTDIR "/vlog/case2.v");
+
+   const error_t expect[] = {
+      { 13, "multiple default statements within a single case statement" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   do_parse_check(V_MODULE);
 
    fail_unless(vlog_parse() == NULL);
 
@@ -920,17 +972,8 @@ START_TEST(test_direct1)
    };
    expect_errors(expect);
 
-   vlog_node_t m1 = vlog_parse();
-   fail_if(m1 == NULL);
-   fail_unless(vlog_kind(m1) == V_MODULE);
-
-   vlog_check(m1);
-
-   vlog_node_t m2 = vlog_parse();
-   fail_if(m2 == NULL);
-   fail_unless(vlog_kind(m2) == V_MODULE);
-
-   vlog_check(m2);
+   for (int i = 0; i < 2; i++)
+      do_parse_only(V_MODULE);
 
    fail_unless(vlog_parse() == NULL);
 
@@ -976,10 +1019,7 @@ START_TEST(test_generate1)
    };
    expect_errors(expect);
 
-   vlog_node_t m1 = vlog_parse();
-   fail_if(m1 == NULL);
-   fail_unless(vlog_kind(m1) == V_MODULE);
-
+   vlog_node_t m1 = do_parse_only(V_MODULE);
    vlog_check(m1);
 
    fail_unless(vlog_parse() == NULL);
@@ -993,23 +1033,40 @@ START_TEST(test_tfcall1)
    input_from_file(TESTDIR "/vlog/tfcall1.sv");
 
    const error_t expect[] = {
-      { 43, "return statement can only be used in a subroutine" },
-      { 40, "return statement in a task cannot have an expression" },
-      { 46, "return statement in a non-void function must have an expression" },
-      {  8, "expected 2 arguments for 'sum' but have 1" },
-      {  9, "expected 2 arguments for 'sum' but have 3" },
-      { 13, "'x4' is not a function" },
-      { 27, "'no_args' is not a task" },
-      { 12, "no visible declaration for 'not_here'" },
+      { 36, "return statement can only be used in a subroutine" },
+      { 39, "'bob' does not match task name 'task3'" },
+      { 42, "'blah' does not match function name 'func2'" },
+      { 11, "no visible declaration for 'not_here'" },
       { -1, NULL }
    };
    expect_errors(expect);
 
-   vlog_node_t m1 = vlog_parse();
-   fail_if(m1 == NULL);
-   fail_unless(vlog_kind(m1) == V_MODULE);
+   do_parse_only(V_MODULE);
 
-   vlog_check(m1);
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
+START_TEST(test_tfcall2)
+{
+   input_from_file(TESTDIR "/vlog/tfcall2.sv");
+
+   const error_t expect[] = {
+      { 40, "return statement in a task cannot have an expression" },
+      { 44, "return statement in a non-void function must have an expression" },
+      { 53, "completely declared port 'z' cannot be declared again in a data "
+        "type declaration" },
+      {  9, "expected 2 arguments for 'sum' but have 1" },
+      { 10, "expected 2 arguments for 'sum' but have 3" },
+      { 13, "'x4' is not a function" },
+      { 27, "'no_args' is not a task" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   do_parse_check(V_MODULE);
 
    fail_unless(vlog_parse() == NULL);
 
@@ -1029,11 +1086,8 @@ START_TEST(test_attr1)
    };
    expect_errors(expect);
 
-   vlog_node_t m1 = vlog_parse();
-   fail_if(m1 == NULL);
-   fail_unless(vlog_kind(m1) == V_MODULE);
-
-   vlog_check(m1);
+   for (int i = 0; i < 2; i++)
+      do_parse_only(V_MODULE);
 
    fail_unless(vlog_parse() == NULL);
 
@@ -1056,11 +1110,7 @@ START_TEST(test_initial1)
    };
    expect_errors(expect);
 
-   vlog_node_t m1 = vlog_parse();
-   fail_if(m1 == NULL);
-   fail_unless(vlog_kind(m1) == V_MODULE);
-
-   vlog_check(m1);
+   do_parse_only(V_MODULE);
 
    fail_unless(vlog_parse() == NULL);
 
@@ -1082,10 +1132,7 @@ START_TEST(test_nets1)
    };
    expect_errors(expect);
 
-   vlog_node_t m1 = vlog_parse();
-   fail_if(m1 == NULL);
-   fail_unless(vlog_kind(m1) == V_MODULE);
-
+   vlog_node_t m1 = do_parse_only(V_MODULE);
    vlog_check(m1);
 
    fail_unless(vlog_parse() == NULL);
@@ -1108,11 +1155,8 @@ START_TEST(test_keywords)
    };
    expect_errors(expect);
 
-   for (int i = 0; i < 2; i++) {
-      vlog_node_t m = vlog_parse();
-      fail_if(m == NULL);
-      fail_unless(vlog_kind(m) == V_MODULE);
-   }
+   for (int i = 0; i < 2; i++)
+      do_parse_only(V_MODULE);
 
    fail_unless(vlog_parse() == NULL);
 
@@ -1126,13 +1170,13 @@ START_TEST(test_error1)
 
    const error_t expect[] = {
       {  2, "no data type declaration for 'foo'" },
+      {  6, "'error1' does not match module name 'error2'" },
       { -1, NULL }
    };
    expect_errors(expect);
 
-   vlog_node_t m = vlog_parse();
-   fail_if(m == NULL);
-   fail_unless(vlog_kind(m) == V_MODULE);
+   for (int i = 0; i < 2; i++)
+      do_parse_only(V_MODULE);
 
    fail_unless(vlog_parse() == NULL);
 
@@ -1162,9 +1206,7 @@ START_TEST(test_integers1)
    };
    expect_errors(expect);
 
-   vlog_node_t m = vlog_parse();
-   fail_if(m == NULL);
-   fail_unless(vlog_kind(m) == V_MODULE);
+   do_parse_only(V_MODULE);
 
    fail_unless(vlog_parse() == NULL);
 
@@ -1214,9 +1256,7 @@ START_TEST(test_label1)
    };
    expect_errors(expect);
 
-   vlog_node_t m = vlog_parse();
-   fail_if(m == NULL);
-   fail_unless(vlog_kind(m) == V_MODULE);
+   do_parse_only(V_MODULE);
 
    fail_unless(vlog_parse() == NULL);
 
@@ -1229,17 +1269,31 @@ START_TEST(test_href1)
    input_from_file(TESTDIR "/vlog/href1.v");
 
    const error_t expect[] = {
-      {  4, "prefix of hierarchical identifier is not an instance" },
-      {  5, "no visible declaration for 'xx'" },
+      {  6, "no visible declaration for 'xx'" },
+      { 13, "no visible declaration for 'yy'" },
       { -1, NULL }
    };
    expect_errors(expect);
 
-   vlog_node_t m = vlog_parse();
-   fail_if(m == NULL);
-   fail_unless(vlog_kind(m) == V_MODULE);
+   do_parse_only(V_MODULE);
 
-   vlog_check(m);
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
+START_TEST(test_href2)
+{
+   input_from_file(TESTDIR "/vlog/href2.v");
+
+   const error_t expect[] = {
+      {  5, "prefix of hierarchical identifier is not an instance" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   do_parse_check(V_MODULE);
 
    fail_unless(vlog_parse() == NULL);
 
@@ -1257,14 +1311,173 @@ START_TEST(test_package1)
    };
    expect_errors(expect);
 
-   vlog_node_t m = vlog_parse();
-   fail_if(m == NULL);
-   fail_unless(vlog_kind(m) == V_PACKAGE);
+   do_parse_only(V_PACKAGE);
+
+   fail_unless(vlog_parse() == NULL);
 
    check_expected_errors();
 }
 END_TEST
 
+START_TEST(test_class1)
+{
+   input_from_file(TESTDIR "/vlog/class1.sv");
+
+   const error_t expect[] = {
+      { 12, "'foo_t' does not match class name 'bar_t'" },
+      { 18, "new class expression must have class type" },
+      { 20, "class 'foo_t' has no field named 'c'" },
+      { 23, "'bob' does not match program name 'main'" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   do_parse_only(V_PROGRAM);
+
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
+START_TEST(test_class2)
+{
+   input_from_file(TESTDIR "/vlog/class2.sv");
+
+   const error_t expect[] = {
+      { 20, "invalid operands for binary expression" },
+      {  0, "have 'int' and 'class'" },
+      { 22, "invalid operands for binary expression" },
+      {  0, "have 'class' and 'class'" },
+      { 23, "value of type 'bar_t' is not compatible with 'foo_t'" },
+      { 24, "value of type 'integral' is not compatible with 'class'" },
+      { 25, "value of type 'bar_t' is not compatible with 'foo_t'" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   do_parse_check(V_PROGRAM);
+
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
+START_TEST(test_class3)
+{
+   input_from_file(TESTDIR "/vlog/class3.sv");
+
+   do_parse_check(V_PACKAGE);
+   do_parse_check(V_MODULE);
+
+   fail_unless(vlog_parse() == NULL);
+
+   fail_if_errors();
+}
+END_TEST
+
+START_TEST(test_namespace1)
+{
+   input_from_file(TESTDIR "/vlog/namespace1.sv");
+
+   do_parse_only(V_MODULE);
+
+   fail_unless(vlog_parse() == NULL);
+
+   fail_if_errors();
+}
+END_TEST
+
+START_TEST(test_real1)
+{
+   input_from_file(TESTDIR "/vlog/real1.v");
+
+   const error_t expect[] = {
+      {  8, "invalid operands for binary expression" },
+      {  9, "invalid operands for binary expression" },
+      { 10, "invalid operands for binary expression" },
+      { 11, "invalid operands for binary expression" },
+      {  0, "have 'real' and 'real'" },
+      { 13, "value of type 'real' cannot be indexed" },
+      { 14, "invalid operands for unary expression" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   do_parse_check(V_MODULE);
+
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
+START_TEST(test_simp1)
+{
+   input_from_file(TESTDIR "/vlog/simp1.v");
+
+   vlog_node_t m = do_parse_check(V_MODULE);
+   vlog_simp(m);
+
+   const int64_t expect[] = { 5, 1, 7, 1, 2 };
+
+   ck_assert_int_eq(ARRAY_LEN(expect), vlog_decls(m));
+
+   for (int i = 0; i < ARRAY_LEN(expect); i++) {
+      vlog_node_t d = vlog_decl(m, i), value = vlog_value(d);
+      if (vlog_kind(value) == V_NUMBER) {
+         number_t n = vlog_number(value);
+         if (number_integer(n) == expect[i])
+            continue;
+      }
+
+      vlog_dump(d, 0);
+      ck_abort_msg("expected constant %"PRIi64, expect[i]);
+   }
+
+   fail_unless(vlog_parse() == NULL);
+
+   fail_if_errors();
+}
+END_TEST
+
+START_TEST(test_lower1)
+{
+   input_from_file(TESTDIR "/vlog/lower1.v");
+
+   run_elab();
+
+   mir_context_t *mc = get_mir();
+
+   {
+      mir_unit_t *mu = mir_get_unit(mc, ident_new("WORK.lower1.assign#3#9"));
+      ck_assert_ptr_nonnull(mu);
+
+      static const mir_match_t bb1[] = {
+         { MIR_OP_VAR_UPREF, ENUM(1) },
+         { MIR_OP_LOAD },
+         { MIR_OP_VAR_UPREF, ENUM(1) },
+         { MIR_OP_LOAD },
+         { MIR_OP_RESOLVED },
+         { MIR_OP_LOAD },
+         { MIR_OP_PACK },
+         { MIR_OP_VAR_UPREF, ENUM(1) },
+         { MIR_OP_LOAD },
+         { MIR_OP_RESOLVED },
+         { MIR_OP_LOAD },
+         { MIR_OP_PACK },
+         { MIR_OP_BINARY, ENUM(MIR_VEC_BIT_AND) },
+         { MIR_OP_UNPACK },
+         { MIR_OP_PUT_DRIVER, NODE(_), CONST(1) },
+         { MIR_OP_WAIT, BLOCK(1) },
+      };
+      mir_match(mu, 1, bb1);
+   }
+
+   fail_if_errors();
+}
+END_TEST
 
 Suite *get_vlog_tests(void)
 {
@@ -1273,7 +1486,8 @@ Suite *get_vlog_tests(void)
    TCase *tc = nvc_unit_test();
    tcase_add_test(tc, test_dff);
    tcase_add_test(tc, test_simple_sem);
-   tcase_add_test(tc, test_ports);
+   tcase_add_test(tc, test_ports1);
+   tcase_add_test(tc, test_ports2);
    tcase_add_test(tc, test_parse1);
    tcase_add_test(tc, test_number1);
    tcase_add_test(tc, test_number2);
@@ -1286,7 +1500,9 @@ Suite *get_vlog_tests(void)
    tcase_add_test(tc, test_udp1);
    tcase_add_test(tc, test_implicit1);
    tcase_add_test(tc, test_struct1);
+   tcase_add_test(tc, test_struct2);
    tcase_add_test(tc, test_enum1);
+   tcase_add_test(tc, test_enum2);
    tcase_add_test(tc, test_union1);
    tcase_add_test(tc, test_param1);
    tcase_add_test(tc, test_param2);
@@ -1296,10 +1512,12 @@ Suite *get_vlog_tests(void)
    tcase_add_test(tc, test_casename);
    tcase_add_test(tc, test_const1);
    tcase_add_test(tc, test_case1);
+   tcase_add_test(tc, test_case2);
    tcase_add_test(tc, test_direct1);
    tcase_add_test(tc, test_string1);
    tcase_add_test(tc, test_generate1);
    tcase_add_test(tc, test_tfcall1);
+   tcase_add_test(tc, test_tfcall2);
    tcase_add_test(tc, test_attr1);
    tcase_add_test(tc, test_initial1);
    tcase_add_test(tc, test_nets1);
@@ -1310,7 +1528,15 @@ Suite *get_vlog_tests(void)
    tcase_add_test(tc, test_pp6);
    tcase_add_test(tc, test_label1);
    tcase_add_test(tc, test_href1);
+   tcase_add_test(tc, test_href2);
    tcase_add_test(tc, test_package1);
+   tcase_add_test(tc, test_class1);
+   tcase_add_test(tc, test_class2);
+   tcase_add_test(tc, test_class3);
+   tcase_add_test(tc, test_namespace1);
+   tcase_add_test(tc, test_real1);
+   tcase_add_test(tc, test_simp1);
+   tcase_add_test(tc, test_lower1);
    suite_add_tcase(s, tc);
 
    return s;
